// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#ifndef TImage_h
#define TImage_h
#include "TBaseObject.h"
#include "TWindow.h"
#include "TString.h"
#include "TArray.h"
#include "Complex.h"
#include "TFile.h"

#ifndef __ImageDataType
#define __ImageDataType
typedef enum {k1ByteSigned=1,k1ByteUnsigned,k2ByteSigned,k2ByteUnsigned,
	k4ByteSigned,k4ByteUnsigned,k4ByteReal,k8ByteReal,k4ByteComplex,k8ByteComplex }
ImageDataType ;
#endif

typedef	enum {kCosineMask,kHanningMask} ImageMaskType ;
typedef	enum {kPixel,kAngstrom,kNanometer,kInversePixel,kInverseAngstrom,kInverseNanometer} CalibrationUnit ;

#define kFFTForward 1 
#define kFFTInverse -1

#define	kUnassigned -99999.
class TImage : public TBaseObject {
	public:
		TImage() : mWidth(0),mHeight(0),mData(nil),mWindow(nil),mOffscreen(nil) {mSignature = 'imag' ;} ;
		TImage(short w,short h,float *data=nil,float angle=90.,ImageDataType type = k4ByteReal) ;
		TImage(TArray *array,float angle=90.) ;
        TImage(TArray &array,Rect *selectionRect = nil) ;
		TImage(const TImage &u,Rect *rect=nil) ;
		TImage(TFile &theFile) ;

		~TImage() ;
		
		bool	InBounds(long x,long y) { if (x >= 0 && x < mWidth && y >= 0 && y < mHeight) return true ; else return false ; } ;
		
		void	SetPixel(long x,long y, float val) { if (InBounds(x,y)) *mData->GetData(x,y) = val ; } ;
		void	SetPixel(float x,float y, float val) { if (InBounds(near(x),near(y))) mData->SetPixel(x,y,val) ; } ;
		float	GetPixel(long x,long y) { if (InBounds(x,y)) return *mData->GetData(x,y) ; else return 0. ; } ;
		float	GetPixel(float x,float y) { if (InBounds(near(x),near(y))) return mData->GetPixel(x,y) ; else return 0. ; } ;

		void 	SetData(float *data,ImageDataType dataType) ;
		float	*GetData(long n=0,long m=0) { return mData->GetData(n,m) ; } ;
		float	GetPixelAmplitude(float x,float y=0.) { return mData->GetPixelAmplitude(x,y) ; } ;
		float	GetPixelPhase(float x,float y=0.) { return mData->GetPixelPhase(x,y) ; } ;
		TArray	*GetDataArray() { return mData ; } ;
		void	SetName(char *text) { mName.SetString(text) ; } ;
		void	Name(char *text) { mName.GetString(text) ; } ;
		void	ZeroStatistics() { mMinValue=kUnassigned;mMaxValue=kUnassigned;mMeanValue=kUnassigned;mStdDev=kUnassigned; } ;
		char*	Name() { return mName.GetString() ; } ;
		void	GetAbbreviatedName(char *) ;	// Returns abbreviated name
		short	Width() { return mWidth; } ;
		short	Height() { return mHeight ; } ;
		float	Gamma() { return mGamma ; } ;
		float	Min() { if (mMinValue == kUnassigned ) return mMinValue=mData->Min(); else return mMinValue ; } ;
		float	Max() { if (mMaxValue == kUnassigned ) return mMaxValue=mData->Max(); else return mMaxValue ; } ;
		float	Min(short &x,short &y) { return mData->Min(x,y) ; } ;
		float	Max(short &x,short &y) { return mData->Max(x,y) ; } ;
		float	Mean() { if (mMeanValue == kUnassigned ) return mMeanValue=mData->Mean(); else return mMeanValue ; } ;
		float	Sum() { return (mWidth*mHeight*Mean()) ; } ;
		float	Stdv() {  if (mStdDev == kUnassigned ) return mStdDev=mData->Stdv(); else return mStdDev ; } ;
		float	VectorLength(){ return mData->VectorLength() ; } ;

		void	SetBrighness(float value) {SetContrastBrightness(mContrast,value) ; } ;
		void	SetContrast(float value) {SetContrastBrightness(value,mBrightness) ; } ;
		void	SetContrastBrightness(float contrast,float brightness) ;
		float	GetContrastValue() { return mContrast ; } ;
		float	GetBrightnessValue() { return mBrightness ; } ;
		void	Draw() ;
		
		
		TWindow	*GetWindow() { return mWindow ; } ;

		float	GetCalibration() { return mCalibration ; } ;
		void	SetCalibration(float calibration) {  mCalibration = calibration ; } ;
		CalibrationUnit	GetCalibrationUnit() { return mCalibrationUnit ; } ; 
		void	SetCalibrationUnit(CalibrationUnit unit) {  mCalibrationUnit = unit ; } ; 
		
		void	FlipHorizontal() ;
		void	FlipVertical() ;
		void	Mask(ImageMaskType maskType) ;
		void	RotateClockWise(float numDegrees) ;
		void	RotateRight() ;
		void	RotateLeft() ;
		void 	ReSize(short w,short h) ;
		void	Scale(float percX,float percY) ;
		void	Blur() ;
		void	Sharpen() ;
		void	Invert() ;
		void	Display(TWindow *window=nil,Rect *rect=nil) ;
		void	Update(Boolean redraw = true) ;
		void	Repeat(int nX,int nY) ;
		void	SetSampling(float dx,float dy) { mSamplingX = dx ; mSamplingY = dy ; } ;
		void	AdjustSampling() ;
		void	AdjustGamma() ;
		void	Print() ;
		void	ConvertToComplex() {ZeroStatistics() ;  mData->MakeComplex() ; mDataIsComplex = true ; } ;
		void	ConvertToReal() { ZeroStatistics() ; mData->MakeReal() ; mDataIsComplex = false ; } ;
		void	SetDataSpace(Boolean isReciprocalSpace) { mIsRealSpace = !isReciprocalSpace ; } ;
		Boolean	IsRealSpace() { return mIsRealSpace ; } ;
		unsigned char	*mRGBData ;
		
		void	SetDisplayLogScale(bool state) { mDisplayOnLogScale = state ; } ;
		

		TImage*	FourierTransform(short direction = kFFTForward) ;
		TImage 	*Extract(float x,float y,short width,short height) ;
		void	FFT() { mData->FFT() ; ZeroStatistics() ; mIsRealSpace = false ; UpdateCalibration() ; 
						if (mRGBData != nil) delete[] mRGBData; mRGBData = nil ;} ; 
		void	IFFT() { mData->IFFT(false) ; ZeroStatistics() ; mIsRealSpace = true ; mDisplayOnLogScale = false ; UpdateCalibration() ; } ; 
		void	fft() { mData->fft() ; ZeroStatistics() ; mIsRealSpace = false ; UpdateCalibration() ; 
						if (mRGBData != nil) delete[] mRGBData; mRGBData = nil ;} ; 
		void	ifft() { mData->ifft(false) ; ZeroStatistics() ; mIsRealSpace = true ; mDisplayOnLogScale = false ; UpdateCalibration() ; } ; 

		void 	Modulus(bool centerImage = false) { mData->Modulus(centerImage) ; ZeroStatistics() ; if (mIsVisible) Update() ;  mIsRealSpace = true ;  mDataIsComplex = false ; } ;
		void 	Intensity(bool centerImage = false) { mData->Intensity(centerImage) ; ZeroStatistics() ; if (mIsVisible) Update() ;  mIsRealSpace = true ;  mDataIsComplex = false ;} ;
		void 	Phase(bool centerImage = false) { mData->Phase(centerImage) ; ZeroStatistics() ; if (mIsVisible) Update() ;  mIsRealSpace = true ;  mDataIsComplex = false ;} ;
		void 	Real(bool centerImage = false) { mData->RealPart(centerImage) ; ZeroStatistics() ; if (mIsVisible) Update() ;  mIsRealSpace = true ; mDataIsComplex = false ;} ;
		void 	Imaginary(bool centerImage = false) { mData->ImaginaryPart(centerImage) ;  ZeroStatistics() ; if (mIsVisible) Update() ; mIsRealSpace = true ;  mDataIsComplex = false ;} ;
		void 	log() { mData->log() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	log10() { mData->log10() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	pow(float x) { mData->pow(x) ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	sqrt() { mData->sqrt() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	sq() { mData->sq() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	exp() { mData->exp() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	sin() { mData->sin() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	cos() { mData->cos() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	tan() { mData->tan() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	asin() { mData->asin() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	acos() { mData->acos() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;
		void 	atan() { mData->atan() ; ZeroStatistics() ; if (mIsVisible) Update() ; } ;

		TImage&	operator=(const float v) ;
		TImage&	operator=(const Cmplx& v) ;
		TImage&	operator=(const TImage& u) ;
		
		TImage&	operator+=(const float v) {*(this->mData) += v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator-=(const float v) {*(this->mData) -= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator*=(const float v) {*(this->mData) *= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator/=(const float v) {*(this->mData) /= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator+=(const Cmplx& v) {*(this->mData) += v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator-=(const Cmplx& v) {*(this->mData) -= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator*=(const Cmplx& v) {*(this->mData) *= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator/=(const Cmplx& v) {*(this->mData) /= v ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator+=(const TImage& u) {*(this->mData) += *(u.mData) ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator-=(const TImage& u) {*(this->mData) -= *(u.mData) ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator*=(const TImage& u) {*(this->mData) *= *(u.mData) ; ZeroStatistics() ; return *this ; } ;
		TImage&	operator/=(const TImage& u) {*(this->mData) /= *(u.mData) ; ZeroStatistics() ; return *this ; } ;

		TImage&	operator^=(const float v) {*(this->mData) ^= v ; ZeroStatistics(); return *this ; } ;

		friend TImage operator*(const float v,const TImage& u) { TImage r = u ; r *= v; return r ; } ;
		friend TImage operator+(const float v,const TImage& u) { TImage r = u ; r += v; return r ; } ;
		friend TImage operator-(const float v,const TImage& u) { TImage r = -1.*u ; r += v; return r ; } ;

		friend TImage operator*(const Cmplx& v,const TImage& u) { TImage r = u ; r *= v; return r ; } ;
		friend TImage operator+(const Cmplx& v,const TImage& u) { TImage r = u ; r += v; return r ; } ;
		friend TImage operator-(const Cmplx& v,const TImage& u) { TImage r = -1.*u ; r += v; return r ; } ;

		friend TImage operator*(const TImage& u,const float v) { TImage r = u ; r *= v; return r ; } ;
		friend TImage operator/(const TImage& u,const float v) { TImage r = u ; r /= v; return r ; } ;
		friend TImage operator+(const TImage& u,const float v) { TImage r = u ; r += v; return r ; } ;
		friend TImage operator-(const TImage& u,const float v) { TImage r = u ; r -= v; return r ; } ;
		friend TImage operator^(const TImage& u,const float v) { TImage r = u ; r ^= v; return r ; } ;

		friend TImage operator*(const TImage& u,const Cmplx& v) { TImage r = u ; r *= v; return r ; } ;
		friend TImage operator/(const TImage& u,const Cmplx& v) { TImage r = u ; r /= v; return r ; } ;
		friend TImage operator+(const TImage& u,const Cmplx& v) { TImage r = u ; r += v; return r ; } ;
		friend TImage operator-(const TImage& u,const Cmplx& v) { TImage r = u ; r -= v; return r ; } ;

		friend TImage operator*(const TImage& u,const TImage& v) { TImage r = u ; r *= v; return r ; } ;
		friend TImage operator/(const TImage& u,const TImage& v) { TImage r = u ; r /= v; return r ; } ;
		friend TImage operator+(const TImage& u,const TImage& v) { TImage r = u ; r += v; return r ; } ;
		friend TImage operator-(const TImage& u,const TImage& v) { TImage r = u ; r -= v; return r ; } ;

		friend TImage Modulus(TImage &u) { TImage r = u ; r.Modulus() ; return r ; } ;
		friend TImage Intensity(TImage &u) { TImage r = u ; r.Intensity() ; return r ; } ;
		friend TImage Phase(TImage &u) { TImage r = u ; r.Phase() ; return r ; } ;
		friend TImage log(TImage &u) { TImage r = u ; r.log() ; return r ; } ;
		friend TImage log10(TImage &u) { TImage r = u ; r.log10() ; return r ; } ;
		friend TImage pow(TImage &u,float x) { TImage r = u ; r.pow(x) ; return r ; } ;
		friend TImage sqrt(TImage &u) { TImage r = u ; r.sqrt() ; return r ; } ;
		friend TImage sq(TImage &u) { TImage r = u ; r.sq() ; return r ; } ;
		friend TImage exp(TImage &u) { TImage r = u ; r.exp() ; return r ; } ;
		friend TImage sin(TImage &u) { TImage r = u ; r.sin() ; return r ; } ;
		friend TImage cos(TImage &u) { TImage r = u ; r.cos() ; return r ; } ;
		friend TImage tan(TImage &u) { TImage r = u ; r.tan() ; return r ; } ;
		friend TImage asin(TImage &u) { TImage r = u ; r.asin() ; return r ; } ;
		friend TImage acos(TImage &u) { TImage r = u ; r.acos() ; return r ; } ;
		friend TImage atan(TImage &u) { TImage r = u ; r.atan() ; return r ; } ;

		friend float InnerProduct(const TImage& u,const TImage& v) { return InnerProduct(*u.mData,*v.mData)  ; } ;

		friend TImage 	FFT(TImage &u) { TImage r = u ; r.mData->fft() ; r.ZeroStatistics() ; return r ; } ; 
		friend TImage 	IFFT(TImage &u) { TImage r = u ; r.mData->ifft() ; r.ZeroStatistics() ; return r ; } ; 
		friend TImage 	FFT(TImage *u) { TImage r = *u ; r.mData->fft() ; r.ZeroStatistics() ; return r ; } ; 
		friend TImage 	IFFT(TImage *u) { TImage r = *u ; r.mData->ifft() ; r.ZeroStatistics() ; return r ; } ; 
		friend TImage 	*WienerFilter(TImage *u) ;
		friend TImage 	*WienerFilter2(TImage *u) ;
		friend TImage	*WienerFiltera(TImage *u) ;
		friend TImage 	*CrossCorrelate(TImage *u,TImage *v) ;
		friend	void	SetName(TImage &u, char *text) { u.mName.SetString(text) ; } ;
		friend	void	GetName(TImage &u,char *text) { u.mName.GetString(text) ; } ;
		friend	void	GetSize(TImage &u,short *w, short *h) { *w = u.mWidth; *h = u.mHeight ; } ;
		friend	float	Gamma(TImage& u) { return u.mGamma ; } ;
		// This should always recompute any min, max, std, etc.
		//friend	float	Min(TImage& u) { return ::Min(*u.mData) ; } ;
		//friend	float	Max(TImage& u) { return ::Max(*u.mData) ; } ;
		//friend	float	Mean(TImage& u) { return ::Mean(*u.mData) ; } ;
		//friend	float	Stdv(TImage& u) { return ::Stdv(*u.mData) ; } ;
		//friend	float	Sum(TImage& u) { return ::Sum(*u.mData) ; } ;
		friend	void	FlipHorizontal(TImage& u) { u.FlipHorizontal() ; } ;
		friend	void	FlipVertical(TImage& u) { u.FlipVertical() ; } ;
		friend	void	Rotate(TImage& u,float angle) { u.RotateClockWise(angle) ; } ;
		friend	void 	ReSize(TImage& u,short w,short h) { u.ReSize(w,h) ; } ;
		friend	void	Scale(TImage& u,float percX,float percY) { u.Scale(percX,percY) ; } ;
		friend	void	DisplayImage(TImage& u) { u.Display() ; } ;
		friend	void	Update(TImage& u) { u.Update() ; } ;
		friend	void	CopyImageToImage(TImage *u,Rect *uRect,TImage *v,Rect *vRect,float multiplicationConstant = 1.) ;

		friend	void	OpenImageFile(FSInfo &theFSInfo) ;

		friend	class	TImageWindow ;

	protected:

		virtual	void	CreateOffscreen() ;
		void	UpdateCalibration() ;
		
		friend			void	DisplayImageInRect(TImage *image, Rect *rect) ;

		short 			mWidth ;
		short 			mHeight ;
		float 			mGamma ;
		float			mContrast ;
		float			mBrightness ;
		float			mMinDisplayLimit ;
		float			mMaxDisplayLimit ;
		unsigned char	*mDisplayData ;
		TString  		mName ;
		Rect			mDisplayRect ;
		TArray*			mData ;
		TOffscreen*		mOffscreen ;
		TWindow*		mWindow ;	// The window that we are associated with

		float 			mMinValue ;
		float 			mMaxValue ;
		float			mMeanValue ;
		float			mStdDev ;
		float			mSamplingX ;
		float			mSamplingY ;
		Boolean 		mIsVisible ;
		Boolean 		mIsDirty ;
		Boolean 		mIsRealSpace ;	// Real Space or Fourier Space
		Boolean			mDataIsComplex ;
		
		Boolean			mDisplayOnLogScale ;
		
		float			mCalibration ;
		CalibrationUnit	mCalibrationUnit ;

} ;

typedef	TImage	*TImageRef ;
TImage *ImportTiffFile(char	*fileName) ;
TImage *ImportRawFile(char	*fileName) ;
void SaveBinaryImage(TImage *image) ;
void SaveGrayscaleTiffImage(TImage *image) ;
void SaveRGBImage(TImage *image) ;

#endif
