      SUBROUTINE CENSCL (X,W,N1,N2,IOPT,ITER,P1TAIL,CENTER,SCALE,IFAIL
     X ,BISCALEIN,S5,Weight)
C
C    *  *  *  *  *  *  *  *  *  * CPR242 *  *  *  *  *  *  *  *  *  *  *
C    CENSCL   WRITTEN BY CHARLES P. REEVE
C    FOR: COMPUTING STATISTICAL ESTIMATES OF THE CENTER AND SCALE
C         OF ENTRIES N1 THROUGH N2 (INCLUSIVE) IN A SET OF DATA.
C         CLASSICAL OR ROBUST ESTIMATES MAY BE SPECIFIED.
C    EXTERNAL SUBPROGRAMS CALLED: SORT1
C    INTERNAL SUBPROGRAMS CALLED: MAD, MEANSD, MEDIAN, vcopyl
C    CURRENT VERSION COMPLETED MARCH 2, 1983
C    *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
C
C    DEFINITION OF PASSED PARAMETERS:
C
C    * X(*) =  INPUT: VECTOR (DIMENSION >= N2) CONTAINING DATA
C             OUTPUT: NO CHANGE
C
C      W(*) = WORKSPACE VECTOR (DIMENSION >= N2)
C
C      * N1 = FIRST ENTRY OF INTEREST IN DATA VECTOR X
C
C      * N2 = LAST ENTRY OF INTEREST IN DATA VECTOR X
C
C    * IOPT = 1 TO COMPUTE MEAN AND STANDARD DEVIATION
C             2 TO COMPUTE MEDIAN AND MEDIAN ABSOLUTE DEVIATION
C             3 TO COMPUTE TRIMMED MEAN AND STANDARD DEVIATION
C             4 TO COMPUTE WINSORIZED MEAN AND STANDARD DEVIATION
C             5 TO COMPUTE M(T) AND S(T)  (REEVE)
C             6 TO COMPUTE M(R) AND S(R)  (REEVE)
C             7 TO COMPUTE BIWEIGHT MEAN AND STD. DEV. (GROSS)
C             8 BIWEIGHT with initial estimate in center
C             9 BIWEIGHT with iterations
C
C    * ITER = MAXIMUM NUMBER OF ITERATIONS TO BE USED IN COMPUTATION
C             OF BIWEIGHT ESTIMATE OF LOCATION.  A POSITIVE VALUE WILL
C             CAUSE A DIAGNOSTIC MESSAGE TO BE PRINTED WHEN 'ITER'
C             ITERATIONS HAVE OCCURRED.  A NEGATIVE VALUE WILL CAUSE
C             NO DIAGNOSTIC MESSAGE TO BE PRINTED.  IN EITHER CASE, THE
C             LATEST COMPUTED VALUE OF THE BIWEIGHT ESTIMATE IS RETURNED
C             WHEN 'ABS(ITER)' ITERATIONS HAVE OCCURRED.  CONVERGENCE TO
C             'X(K)' WILL OCCUR EARLIER ON ITERATION 'K' IF
C
C               ABS( X(K)-X(K-1) )  <  0.0001 * ABS( X(K) ) + .000001
C
C
C  * P1TAIL = PROPORTION OF ORDERED DATA AT EACH EXTREME TO BE
C             CONSIDERED THE TAIL REGION (USED IF IOPT=3 OR 4)
C
C    CENTER = MEAN                                    IF IOPT=1
C             MEDIAN                                  IF IOPT=2
C             TRIMMED MEAN                            IF IOPT=3
C             WINSORIZED MEAN                         IF IOPT=4
C             M(T)  (REEVE)                           IF IOPT=5
C             M(R)  (REEVE)                           IF IOPT=6
C             BIWEIGHT MEAN                           IF IOPT=7
C             ITERATIVE BIWEIGHT MEAN                 IF IOPT=9
C
C     SCALE = STANDARD DEVIATION                      IF IOPT=1
C             MEDIAN ABSOLUTE DEVIATION               IF IOPT=2
C             TRIMMED STANDARD DEVIATION              IF IOPT=3
C             WINSORIZED STANDARD DEVIATION           IF IOPT=4
C             S(T)  (REEVE)                           IF IOPT=5
C             S(R)  (REEVE)                           IF IOPT=6
C             BIWEIGHT STD. DEV.  (GROSS)             IF IOPT=7
C             ITERATIVE BIWEIGHT STD DEV              IF IOPT=9
C
C    * DENOTES VARIABLES REQUIRING INPUT VALUES
C
C    *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION X(*),W(*),WOLD(2000),weight(*)
      COMMON /TWEAK/ SCALT
      parameter (toofew = 0.8D0)
      BISCALE=BISCALEIN
      IF(BISCALE .LT. 1.D0) BISCALE=6.D0
C
C--- COMPUTE NUMBER OF DATA POINTS USED
C
      N = N2-N1+1
C
C--- CHECK FOR AT LEAST FOUR (4) DATA POINTS -- OTHERWISE, STOP!
C
      IFAIL=0
      IF (N.GE.4) GO TO 10
      IFAIL=-N
      RETURN
C      PRINT *,' *** NUMBER OF DATA POINTS IS < 4'
C      PRINT *,' *** EXECUTION STOPPED IN SUBROUTINE CENSCL'
C      STOP
C
C--- SET ACCIDENTAL NEGATIVE TAIL PROPORTION TO ZERO
C
   10 P1TAIL = MAX(P1TAIL,0.D0)
C
C--- BRANCH TO CLASSICAL OR ROBUST ESTIMATES
C
      GO TO (20,30,40,50,70,70,130,130,230,330), IOPT
C
C****************************************************************
C*** COMPUTE CLASSICAL ESTIMATES: MEAN AND STANDARD DEVIATION ***
C****************************************************************
C
   20 CALL MEANSD (X,N1,N2,CENTER,SCALE)
      RETURN
C
C**********************************************************************
C*** COMPUTE ROBUST ESTIMATES: MEDIAN AND MEDIAN ABSOLUTE DEVIATION ***
C**********************************************************************
C
   30 CALL vcopyl (X,W,N1,N2)
      CALL MEDIAN (W,N1,N2,CENTER)
      CALL MAD (W,N1,N2,CENTER,SCALE)
      RETURN
C
C*********************************************************************
C*** COMPUTE ROBUST ESTIMATES: TRIMMED MEAN AND STANDARD DEVIATION ***
C*********************************************************************
C
   40 CALL vcopyl (X,W,N1,N2)
      CALL SORT1 (W,N1,N2)
      NTRIM = MIN0((N-1)/2,NINT(P1TAIL*DBLE(N)))
      NLO = N1+NTRIM
      NHI = N2-NTRIM
      CALL MEANSD (W,NLO,NHI,CENTER,SCALE)
      RETURN
C
C***************************************************************
C*** COMPUTE ROBUST ESTIMATES: WINSORIZED MEAN AND STD. DEV. ***
C***************************************************************
C
   50 CALL vcopyl (X,W,N1,N2)
      CALL SORT1 (W,N1,N2)
      NTRIM = MIN0((N-1)/2,NINT(P1TAIL*DBLE(N)))
      DO 60 I = 1, NTRIM
         W(N1+I-1) = W(N1+NTRIM)
         W(N2-I+1) = W(N2-NTRIM)
   60 CONTINUE
      CALL MEANSD (W,N1,N2,CENTER,SCALE)
      RETURN
C
C****************************************************************
C*** COMPUTE ROBUST ESTIMATES: M(T) AND S(T) OR M(R) AND S(R) ***
C****************************************************************
C
   70 CALL vcopyl (X,W,N1,N2)
      CALL SORT1 (W,N1,N2)
C
C--- SET CONSTANT WHICH RELATES SHORTEST INTERVAL TO STANDARD
C--- DEVIATION OF THE NORMAL DISTRIBUTION
C
      CNORM = 1.348979D0
C
C--- SET REJECTION LIMITS AS A MULTIPLE OF THE STANDARD DEVIATION
C
      CRHO = 3.0D0
      RMIN = 1D30
      L = N/2
      NR = (N-1)/2
      B = DBLE(L-NR+1)
      A = 4.D0-B
C
C--- FIND THE SHORTEST INTERVAL COVERING HALF THE POINTS
C
      K1 = 1
      DO 80 I = 1, NR
         J = N1+I-1
         K = J+L
         R = (A*(W(K+1)-W(J))+B*(W(K)-W(J+1)))/4.D0
         IF (R.GE.RMIN) GO TO 80
         RMIN = R
         K1 = J
   80 CONTINUE
      K2 = K1+1
      K3 = K1+L
      K4 = K3+1
      CENTER = (A*(W(K1)+W(K4))+B*(W(K2)+W(K3)))/8.D0
      SCALE = RMIN/CNORM
C
C--- RETURN M(R) AND S(R) IF IOPT=6
C
      IF (IOPT.EQ.6) RETURN
      BOUND = CRHO*SCALE
C
C--- TRIM OUTLYING OBSERVATIONS AT LOWER END
C
      NLO = N1
   90 IF (CENTER-W(NLO).LE.BOUND) GO TO 100
      NLO = NLO+1
      GO TO 90
C
  100 NHI = N2
C
C--- TRIM OUTLYING OBSERVATIONS AT UPPER END
C
  110 IF (W(NHI)-CENTER.LE.BOUND) GO TO 120
      NHI = NHI-1
      GO TO 110
C
  120 CALL MEANSD (W,NLO,NHI,CENTER,SCALE)
C
C--- RETURN M(T) AND S(T) FOR IOPT=5
C
      RETURN
C
C**********************************************************************
C*** COMPUTE ROBUST ESTIMATES: BIWEIGHT MEAN AND STANDARD DEVIATION ***
C**********************************************************************
C--- DEFINE CONVERGENCE CONSTANTS
C
  130 C1 = 1D-4
      C2 = 1D-6
      ITERA = IABS(ITER)
C
C--- COUNT NUMBER OF ITERATIONS FOR BIWEIGHT MEAN COMPUTATION
C
      KOUNT = 0
      CALL vcopyl (X,W,N1,N2)
      DO I = N1,N2
         WOLD(I)=1.
      ENDDO
      IF(IOPT .EQ. 8)goto 140
      CALL MEDIAN (W,N1,N2,CENTER)
      CENTER = 0.
      DO I = N1, N2
         CENTER = CENTER + W(I)
      ENDDO
      CENTER = CENTER/(N2-N1+1)
  140 CALL MAD (W,N1,N2,CENTER,SCALE)
      S1 = 0.D0
      S2 = 0.D0

      DO 150 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
C        Changed here to reduce speed -- less oscillations
         IF (ABS(U).LT.1.0) THEN
            U = (WOLD(I)+(1.D0-U*U)**2  )*0.5
         ELSE
            U = WOLD(I)*0.5
         ENDIF
         WOLD(I)=U
         S1 = S1+U*W(I)
         S2 = S2+U
  150 CONTINUE
      COLD = CENTER
      CENTER = S1/S2
C     Changed exit to do at least 5 iterations
      IF ( (KOUNT .GE. 5) .AND.
     X (ABS(CENTER-COLD).LT.C1*ABS(CENTER)+C2) ) GO TO 160
      KOUNT = KOUNT+1
C      write(6,*)kount,itera,iter ,'No iter'
      IF (KOUNT.LE.ITERA) GO TO 140
      IF (ITER.LT.0) GO TO 160
      N = N2-N1+1
C      WRITE (6,1010) ITER,COLD,CENTER,N
  160 S1 = 0.D0
C      S2 = 0.D0
      S3 = 0.D0
      S4 = 0.D0
      S5 = 0.D0
      DO 170 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE) ! Was 9.0 here which is odd
         W(I)=0
         IF (ABS(U).GE.1.0) GO TO 170
         V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
         S1 = S1+(X(I)-CENTER)**2*V**4     ! Sum (Err*Wt)**2
C         S2 = S2+V*(1.D0-5.D0*U*U)         ! Not sure what this is
         S3 = S3+abs(X(I)-CENTER)*V*V      ! Sum abs(Err*Wt)
         S4 = S4 + V*V*V*V                 ! Sum Wt**2
         S5 = S5 + V*V                     ! Sum Wt
         W(I)=V*V
  170 CONTINUE
C      SCALE = SQRT(DBLE(N)*S1/(S2*(S2-1.D0)))
C     Other formula
C     Error of the mean as an estimate
C      SCALT = scale/sqrt(S5-1.D0)
C     Odd value from original code
C      SCALE = SQRT(DBLE(N)*S1)/abs(S2)
C     Simpler versions
      SCALE = sqrt(S1/S4)
      SCALT= S3/S5
C     write(88,88)N2-N1,S5,sqrt(S4),S1,S3
88    format(i4,4E12.5)
C      if(scale .lt. scalt) scale=scalt
C      write(6,*)'Done'
      RETURN
 1010 FORMAT (1X,'*** ',I3,
     *   ' ITERATION LIMIT ON BIWEIGHT ***    OLD CENTER =',G12.5,3X,
     *   'NEW CENTER =',G12.5,3X,'SAMPLE SIZE =',I6)
C
C     Additional code for iterative version
C**********************************************************************
C*** COMPUTE ROBUST ESTIMATES: BIWEIGHT MEAN AND STANDARD DEVIATION ***
C*** Iterative
C**********************************************************************
C--- DEFINE CONVERGENCE CONSTANTS
C
  230 C1 = 1D-4
      C2 = 1D-6
      ITERA = IABS(ITER)
C
C--- COUNT NUMBER OF ITERATIONS FOR BIWEIGHT MEAN COMPUTATION
C
      KOUNT = 0
      CALL vcopyl (X,W,N1,N2)
      CALL MEDIAN (W,N1,N2,CENTER)
  240 CALL MAD (W,N1,N2,CENTER,SCALE)
      DO I = N1,N2
         WOLD(I)=1.
      ENDDO
  241 S1 = 0.D0
      S2 = 0.D0

      DO 250 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).LT.1.0) THEN
            U = (WOLD(I)+(1.D0-U*U)**2  )*0.5
         ELSE
            U = WOLD(I)*0.5
         ENDIF
         WOLD(I)=U
         S1 = S1+U*X(I)
         S2 = S2+U
  250 CONTINUE
      COLD = CENTER
      CENTER = S1/S2
C     Find the new scale, iteratively
      S3 = 0.D0
      S5 = 0.D0
      DO I = N1, N2
            V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
            S3 = S3+abs(X(I)-CENTER)*WOLD(I)  ! Sum abs(Err*Wt)
            S5 = S5 + WOLD(I)                 ! Sum Wt
      ENDDO
      SCALE = (SCALE + S3/S5)*0.5
C     Changed exit to do at least 5 iterations
      IF ( (KOUNT .GE. 5) .AND.
     X (ABS(CENTER-COLD).LT.C1*ABS(CENTER)+C2) ) GO TO 260
      KOUNT = KOUNT+1
C      write(6,*)kount,itera,'Iter',iter
      IF (KOUNT.LE.ITERA) GO TO 241
      IF (ITER.LT.0) GO TO 260
      N = N2-N1+1
C      WRITE (6,1010) ITER,COLD,CENTER,N
  260 S1 = 0.D0
      S3 = 0.D0
      S4 = 0.D0
      S5 = 0.D0
      DO 270 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).GE.1.0) GO TO 270
         V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
         S1 = S1+(X(I)-CENTER)**2*V**4     ! Sum (Err*Wt)**2
         S3 = S3+abs(X(I)-CENTER)*V*V      ! Sum abs(Err*Wt)
         S4 = S4 + V*V*V*V                 ! Sum Wt**2
         S5 = S5 + V*V                     ! Sum Wt
  270 CONTINUE
C     Did we fail?
      if(S5 .lt. toofew*(N2-N1+1))then
            BISCALE=BISCALE*1.05
C            write(6,*)Biscale,scale,center,S5
            goto 230
      else if(S5 .gt. 0.95*(N2-N1+1))then
            BISCALE=BISCALE/1.05
            goto 230
      endif
C     Simpler versions

      DO I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).LT.1.0) THEN
            V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
            W(I)=V*V
         ELSE
            W(I)=0
         ENDIF
      ENDDO
C     write(88,88)N2-N1,S5,sqrt(S4),S1,S3
      SCALE = sqrt(S1/S4)
      SCALT= S3/S5
C      write(6,*)'Done'
      RETURN
C
C     Additional code for iterative version, starting from scales
C**********************************************************************
C*** COMPUTE ROBUST ESTIMATES: BIWEIGHT MEAN AND STANDARD DEVIATION ***
C*** Iterative
C**********************************************************************
C--- DEFINE CONVERGENCE CONSTANTS
C
  330 C1 = 1D-4
      C2 = 1D-6
      ITERA = IABS(ITER)
C
C--- COUNT NUMBER OF ITERATIONS FOR BIWEIGHT MEAN COMPUTATION
C
      KOUNT = 0
      CALL vcopyl (X,W,N1,N2)
      CALL MEDIAN (W,N1,N2,CENTER)
  340 CALL MAD (W,N1,N2,CENTER,SCALE)
      wav = 0.
      DO I = N1,N2
         WOLD(I)=weight(i)
         wav = wav + weight(i)
      ENDDO
      wav = wav/(N2-n1+1)
  341 S1 = 0.D0
      S2 = 0.D0

      DO 350 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).LT.1.0) THEN
            U = (WOLD(I)+weight(i)*(1.D0-U*U)**2  )*0.5
         ELSE
            U = WOLD(I)*0.5
         ENDIF
         WOLD(I)=U
         S1 = S1+U*X(I)
         S2 = S2+U
  350 CONTINUE
      COLD = CENTER
      CENTER = S1/S2
C     Find the new scale, iteratively
      S3 = 0.D0
      S5 = 0.D0
      DO I = N1, N2
C            V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
            S3 = S3+abs(X(I)-CENTER)*WOLD(I)  ! Sum abs(Err*Wt)
            S5 = S5 + WOLD(I)                 ! Sum Wt
      ENDDO
      SCALE = (SCALE + S3/S5)*0.5
C     Changed exit to do at least 5 iterations
      IF ( (KOUNT .GE. 5) .AND.
     X (ABS(CENTER-COLD).LT.C1*ABS(CENTER)+C2) ) GO TO 360
      KOUNT = KOUNT+1
C      write(6,*)kount,itera
      IF (KOUNT.LE.ITERA) GO TO 341
      IF (ITER.LT.0) GO TO 360
      N = N2-N1+1
C      WRITE (6,1010) ITER,COLD,CENTER,N
  360 S1 = 0.D0
      S3 = 0.D0
      S4 = 0.D0
      S5 = 0.D0
      DO 370 I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).GE.1.0) GO TO 370
         V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
         VV = V*V*weight(i)
         V4 = VV*VV
         S1 = S1+(X(I)-CENTER)**2*V4     ! Sum (Err*Wt)**2
         S3 = S3+abs(X(I)-CENTER)*VV      ! Sum abs(Err*Wt)
         S4 = S4 + V4                ! Sum Wt**2
         S5 = S5 + VV                    ! Sum Wt
  370 CONTINUE
C     Did we fail?
      if(S5/wav .lt. toofew*(N2-N1+1))then
            BISCALE=BISCALE*1.05
C            write(6,*)'Looping ',biscale,scale,center
            goto 330
      else if(S5 .gt. 0.95*(N2-N1+1))then
            BISCALE=BISCALE/1.05
            goto 330
      endif
C     Simpler versions

      DO I = N1, N2
         U = (X(I)-CENTER)/(BISCALE*SCALE)
         IF (ABS(U).LT.1.0) THEN
            V = 1.D0-U*U                      ! Note, V**2 is the Biweight used
            W(I)=V*V
         ELSE
            W(I)=0
         ENDIF
      ENDDO
C     write(88,88)N2-N1,S5,sqrt(S4),S1,S3
      SCALE = sqrt(S1/S4)
      SCALT= S3/S5
      S5=S5/wav
C      write(6,*)'Done'
      RETURN
C
      END
C
C
      SUBROUTINE MEANSD (X,NLO,NHI,AMEAN,SD)
C
C--- COMPUTE CLASSICAL MEAN AND STANDARD DEVIATION OF DATA BETWEEN
C--- POSITIONS X(NLO) AND X(NHI) INCLUSIVE
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION X(*)
      AMEAN = 0.D9
      DO 10 I = NLO, NHI
         AMEAN = AMEAN+X(I)
   10 CONTINUE
      AMEAN = AMEAN/DBLE(NHI-NLO+1)
      SD = 0.D0
      DO 20 I = NLO, NHI
         SD = SD+(X(I)-AMEAN)**2
   20 CONTINUE
      SD = SQRT(SD/DBLE(NHI-NLO))
      RETURN
C
      END
C
C
      SUBROUTINE vcopyl (A,B,NLO,NHI)
      IMPLICIT REAL*8 (A-H,O-Z)
C
C--- COPY SEGMENT OF VECTOR A INTO SAME SEGMENT OF VECTOR B
C
      DIMENSION A(*),B(*)
      DO 10 I = NLO, NHI
         B(I) = A(I)
   10 CONTINUE
      RETURN
C
      END
C
C
      SUBROUTINE MEDIAN (X,NLO,NHI,AMED)
C
C--- COMPUTE MEDIAN OF DATA BETWEEN POSITIONS X(NLO) AND X(NHI)
C--- INCLUSIVE.  DATA IN THIS REGION OF VECTOR X ARE ALTERED.
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION X(*)
C
C--- SORT REGION OF INTEREST IN VECTOR X
C
      CALL SORT1 (X,NLO,NHI)
C
C--- COMPUTE MEDIAN
C
      I = (NLO+NHI)/2
      J = (NLO+NHI+1)/2
      AMED = (X(I)+X(J))/2.D0
      RETURN
C
      END
C
C
      SUBROUTINE MAD (X,NLO,NHI,C,AMAD)
C
C--- COMPUTE MEDIAN ABSOLUTE DEVIATION (MAD) OF X FROM C, AN ESTIMATE
C--- OF THE CENTER OF DATA, BETWEEN X(NLO) AND X(NHI) INCLUSIVE.  VECTOR
C--- X IS EXPECTED TO BE SORTED AND IS UNCHANGED BY THIS SUBROUTINE.
C--- NOTE: IF THE NUMBER OF ENTRIES OF INTEREST, N, IS EVEN THE MAD IS
C--- THE N/2 LARGEST DEVIATION, A SLIGHT OVERESTIMATE.
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION X(*)
      MLO = NLO
      MHI = NHI
      K = (MHI-MLO)/2
      DO 20 I = 1, K
         IF (X(MHI)+X(MLO).GT.2.D0*C) GO TO 10
         MLO = MLO+1
         GO TO 20
C
   10    MHI = MHI-1
   20 CONTINUE
      AMAD = MAX(ABS(X(MHI)-C),ABS(X(MLO)-C))
      RETURN
C
      END
      SUBROUTINE SORT1 (X,M,N)
C
C-----------------------------------------------------------------------
C   SORT1   OBTAINED BY CHARLES P. REEVE FROM DR. D.A. ZAHN AT THE
C           FLORIDA STATE UNIVERSITY, TALLAHASSEE, FLORIDA UNDER THE
C           NAME 'FTASORT'.  SEVERAL MINOR CORRECTIONS WERE MADE TO
C           THE ORIGINAL VERSION AFTER LINE NUMBER 90.
C
C   FOR: SORTING THE SEGMENT OF A DBLE ARRAY BETWEEN ENTRIES M AND N
C        FROM SMALLEST TO LARGEST.  THE ARRAYS LP(K) AND LQ(K) PERMIT
C        SORTING UP TO 2**(K+1)-1 ELEMENTS, THUS K=25 PERMITS SORTING
C        UP TO 67,108,863 ELEMENTS.
C
C    SUBROUTINES CALLED: -NONE-
C
C    CURRENT VERSION COMPLETED JANUARY 30, 1978
C-----------------------------------------------------------------------
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION LP(25),LQ(25),X(*)
      IF (M.GE.N) RETURN
      L = N
      I = 1
      KM = M
      KN = N
      IF (KM.GE.KN) GO TO 70
   10 J = KM
      K = (KN+KM)/2
      A = X(K)
      IF (X(KM).LE.A) GO TO 20
      X(K) = X(KM)
      X(KM) = A
      A = X(K)
   20 L = KN
      IF (X(KN).GE.A) GO TO 40
      X(K) = X(KN)
      X(KN) = A
      A = X(K)
      IF (X(KM).LE.A) GO TO 40
      X(K) = X(KM)
      X(KM) = A
      A = X(K)
      GO TO 40
   30 X(L) = X(J)
      X(J) = B
   40 L = L-1
      IF (X(L).GT.A) GO TO 40
      B = X(L)
   50 J = J+1
      IF (X(J).LT.A) GO TO 50
      IF (J.LE.L) GO TO 30
      IF (L-KM.LE.KN-J) GO TO 60
      LP(I) = KM
      LQ(I) = L
      KM = J
      I = I+1
      GO TO 80
   60 LP(I) = J
      LQ(I) = KN
      KN = L
      I = I+1
      GO TO 80
   70 I = I-1
      IF (I.EQ.0) RETURN
      KM = LP(I)
      KN = LQ(I)
   80 IF (KN-KM.GE.11) GO TO 10
      KM = KM-1
   90 KM = KM+1
      IF (KM.EQ.KN) GO TO 70
      A = X(KM+1)
      IF (X(KM).LE.A) GO TO 90
      J = KM
  100 X(J+1) = X(J)
      J = J-1
      IF (J.EQ.M-1) GO TO 110
      IF (A.LT.X(J)) GO TO 100
  110 X(J+1) = A
      GO TO 90
      END
