/*
 * inv_mat.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "complex.h"
#include "cmatutil.h"
#include "vecmat.h"
#include "inv_mat.h"

void print_complexmat(int n, complex **a);

int inv_c_mat( int n, double eps, complex **a)
{
  /*
   * Inverse matrix
   *   The LU decomposition method
   *   written by KT (ref. MINVC in NUMPAC)
   */
  int *mx;
  int i,j,k,m;
  double am,aa;
  complex s,w,p;
  
  mx = new_vector(int, n);	chk_vector(mx);
  if( (n < 1) || (eps <= 0.0) ){
    printf("$$ (inv_c_mat) invalid n or epsilon.\n");
    free(mx); return 30000;
  }
  
  /* LU decomposition */
  for(j=0; j<n; j++){	/* 1: */
    if(j != 0){		/* 2: */
      for(i=0; i<=j-1; i++){
        m = mx[i];
        s = a[m][j]; a[m][j] = a[i][j];
        if(i!=0){
          for(k=0; k<=i-1; k++){
            s = c_add( s, c_mul(a[i][k], a[k][j]) );
          }
        }
        a[i][j] = s;
      }
    }
    
    am = 0.0;
    for(i=j; i<n; i++){	/* 3: */
      s = a[i][j];
      if(j!=0){
        for(k=0; k<=j-1; k++){
          s = c_add( s, c_mul(a[i][k], a[k][j]) );
        }
        a[i][j] = s;
      }
      aa = c_abs(s);
      if (aa > am){
        am = aa; m = i;
      }
    }
    
    if(am < eps){
      printf("$$ Anomalous termination in inv_mat.\n");
      free_vector(mx); return j;
    }
    
    mx[j] = m;
    
    if(m!=j){		/* 4: */
      for(k=0; k<=j; k++){
        w = a[m][k];			/* swap */
        a[m][k] = a[j][k];
        a[j][k] = w;
      }
    }
    
    if(j == n-1){
      goto IL;
    }
    
    /* 5: */
    w = c_mul( c_conv(-1.0, 0.0), a[j][j] );
    for(i=j+1; i<n; i++){
      a[i][j] = c_div( a[i][j], w );
    }
  }
  
IL:	
  /* Inplace inversion of L compornent */
  if( n>2 ){
    for(i=2; i<n; i++){
      for(j=0; j<=i-2; j++){
        s = a[i][j];
        for(k=j+1; k<=i-1; k++){
          s = c_add( s, c_mul(a[i][k], a[k][j]) );
        }
        a[i][j] = s;
      }
    }
  }
  
  /* Inplace inversion of U compornent */
  a[0][0] = c_div( c_conv(1.0, 0.0), a[0][0] );
  if(n == 1){
    free_vector(mx); return 0;   /* Normal termination */
  }
  for(j=1; j<n; j++){
    a[j][j] = c_div( c_conv(1.0, 0.0), a[j][j] );
    p = c_mul( c_conv(-1.0, 0.0), a[j][j] );
    for(i=0; i<=j-1; i++){
      s = c_conv(0.0, 0.0);
      for(k=i; k<=j-1; k++){
        s = c_add( s, c_mul(a[i][k], a[k][j]) );
      }
      a[i][j] = c_mul(s, p);
    }
  }
  
  /* Inplace multiplication of L and U components */
  for(j=0; j<n-1; j++){
    for(i=0; i<=j; i++){
      s = a[i][j];
      for(k=j+1; k<n; k++){
        s = c_add( s, c_mul(a[i][k], a[k][j]) );
      }
      a[i][j] = s;
    }
    for(i=j+1; i<n; i++){
      s = c_conv(0.0, 0.0);
      for(k=i; k<n; k++){
        s = c_add( s, c_mul(a[i][k], a[k][j]) );
      }
      a[i][j] = s;
    }
  }
  
  /* Interchange of columns */
  for(j=n-2; j>=0; j--){
    m = mx[j];
    if(m != j){
      for(i=0; i<n; i++){
        w = a[i][m];		/* swap */
        a[i][m] = a[i][j];
        a[i][j] = w;
      }
    }
  }
  free_vector(mx); return 0;   /* Normal termination */
}


void inverse( int n, complex **a, complex *b, double eps, int *ierr )
{
  /*
   *  Calculate the first row of c-inverse by Gaussian elimination
   *   and back substitution
   *  *ierr == 0 : normal termination
   */
  complex t, zero;
  int		i, j, k, m, *ipvt, kp1, jpvt, nm1, kb;
  double	amx;
  
  ipvt = new_vector( int, n );  chk_vector( ipvt );
  
  *ierr = 0;	amx = 0.0;
  zero.re = 0.0;	zero.im = 0.0;
  
  if ( n != 1 ) {
    /*
   pivoting for the first row
     */
    for( j=0; j<n; j++ ) {
      if ( c_abs( a[0][j] ) > amx ) {
        amx = c_abs( a[0][j] );
        jpvt = j;
      }
    }
    for( i=0; i<n; i++ ) {
      t = a[i][jpvt];
      a[i][jpvt] = a[i][n-1];
      a[i][n-1] = t;
    }
    /*
   interchange first row and last row
     */
    for( i=0; i<n; i++ ) {
      t = a[0][i];
      a[0][i] = a[n-1][i];
      a[n-1][i] = t;
    }
    
    /*
   Gaussian elimination with partial pivoting
     */
    nm1 = n - 1;
    
    for( k=0; k<nm1; k++ ) {
      kp1 = k + 1;
      m = k;
      if ( kp1 == nm1 || k == nm1 ) { ; }
      else{
        for( i=kp1; i<nm1; i++ ) {
          if( c_abs( a[i][k] ) > c_abs( a[m][k] ) ) {
            m = i;
          }
        }
      }
      
      ipvt[k] = m;
      t = a[m][k];
      a[m][k] = a[k][k];
      a[k][k] = t;
      if ( c_abs( t ) <= eps ) { break; }
      for( i=kp1; i<n; i++ ) {
        a[i][k] = c_div( c_mul(c_conv(-1.0,0.0), a[i][k]), t );
      }
      
      for( j=kp1; j<n; j++ ) {
        t = a[m][j];
        a[m][j] = a[k][j];
        a[k][j] = t;
        if ( t.re == 0.0 && t.im == 0.0 ) { ; }
        else{
          for( i=kp1; i<n; i++ ) {
            a[i][j] = c_add( a[i][j], c_mul( a[i][k], t ) );
          }
        }
      }
    }
  }
  
  if ( c_abs( a[n-1][n-1] ) <= eps ) { *ierr = 30000; }
  
  /*
   back substitution to get x
   */
  b[n-1] = c_div( c_conv(1.0,0.0), a[n-1][n-1] );
  
  for( kb=1; kb<n; kb++ ) {
    k = n - kb - 1;
    kp1 = k + 1;
    b[k] = zero;
    for( j=kp1; j<n; j++ ) {
      b[k] = c_sub( b[k], c_mul( a[k][j], b[j] ) );
    }
    b[k] = c_div( b[k], a[k][k] );
  }
  
  t = b[jpvt];
  b[jpvt] = b[n - 1];
  b[n - 1] = t;
  
  free_vector( ipvt );
  return;
}
