/*
 *
 * mbfunc.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "mbconfig.h"
#include "fitting.h"
#include "eig_mat.h"
#include "cg.h"
#include "inv_mat.h"
#include "perturb.h"
#include "data_out.h"
#include "mb_mpi.h"
#include "tiff.h"
#define BUFLEN  256
extern DEBUG_INFO dbg;
extern EPS_CTRL	epsilon;

void print_complexvec(int n, complex *v);
void print_complexmat(int n, complex **a);

double exct_err(
  double k_z, double abs_incidence, double abs_k,
  HKL g, R_HKL k_xy, UVW incidence, CELL_PAR bv
  );
double tilting_factor(
  double k_z,  double abs_surf_norm,
  double cosfai_inc_surf,
  HKL g, R_HKL k_xy,  UVW surf_norm
  );
double calc_bs_density(
  int nth_branch, R_UVW rx, int num_exact, HKL reflex[],
  double *beta_ge, complex *eigen_value, complex **eigen_vec
  );


void potential_matrix_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex **vg
  )
{
  /*
   * Crystal potential matrix (volt unit)
   *   (with absorption and no relativistic correction)
   *
   *                    [j]
   *           0      1                 n-1
   *     0( a[0][0] a[0][1]              -     )
   *     1( a[1][0] a[1][1]              -     )
   *      (   -          -               -     )
   *  [i] (   -             -            -     )
   *      (   -                -         -     )
   *      (   -                  -       -     )
   *   n-1( a[n-1][0]             a[n-1][n-1]  )
   *
   */

  int i, j, num_rf;
  HKL g, *g_table;
  complex fg, *fg_table;
  int ct_fg_table, g_table_size;

  constraint_csd(csd);

  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);
  chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);
  chk_vector(g_table);

  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    fg = lofg_to_fg(
      csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
      csd->num_sym_op, csd->sym_op,
      csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, kvolts, bv
      );
    stock_fg_tbl_sym(
      csd->g_lofg[i], fg, &ct_fg_table,
      g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }

  num_rf = num_exact_rf + num_bethe_rf;
  for(i=0; i<num_rf; i++){
    progress_indicator(i, num_rf);
    for(j=0; j<num_rf; j++){
      g.h = reflex[i].h - reflex[j].h;
      g.k = reflex[i].k - reflex[j].k;
      g.l = reflex[i].l - reflex[j].l;
      /* modified on 2001.06.14 */
      fg = structure_factor_ext_tbl(
        g, csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->xdyl, csd->ftbl,
        kvolts, bv, csd->lattice_type,
        &ct_fg_table, g_table_size, g_table, fg_table
        );
      vg[i][j] = fg_to_vg(fg, bv);
    }
  }
  mb_fprintf(stdout, "100%%\n");

  free_vector(fg_table);
  free_vector(g_table);
}


void vgv_fxgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv,
  complex *vgv, complex *fxgv
  )
{
  /*
   * Calculation of
   *  Fourier coef. of crystal potential vgv
   *  and X-ray structure factor fxgv
   */
  int  i, num_rf, ng, prog;
  int  ct_fg_table, g_table_size;
  double  sq_s;
  HKL  g, *g_table;
  complex  fg, *fg_table, zgv;
  
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);  chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);  chk_vector(g_table);
  
  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    fg = lofg_to_fg(
      csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
      csd->num_sym_op, csd->sym_op,
      csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, kvolts, bv
      );
    stock_fg_tbl_sym(
      csd->g_lofg[i], fg, &ct_fg_table,
      g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }
  num_rf = num_exact_rf + num_bethe_rf;

  for(i=0; i<num_rf; i++){
    progress_indicator(i, num_rf);
    g = reflex[i];
    if ( ( ng = search_g_table( ct_fg_table, reflex[i], g_table) ) >= 0 ){
      /* found in the table */
      fg = fg_table[ng];
    }
    else{ /* not found in the table */
      fg = calc_Fg_ela(
        g, csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->xdyl, csd->ftbl,
        kvolts, bv
        );
      stock_fg_tbl_sym(
        g, fg, &ct_fg_table, g_table_size, g_table, fg_table,
        csd->num_sym_op, csd->sym_op
        );
    }
    vgv[i] = fg_to_vg(fg, bv);  /* potential V */
    zgv = calc_z_factor(
      g, csd->num_sym_op, csd->sym_op,
      csd->n_atom, csd->atm, csd->xdyl, kvolts, bv, csd->lattice_type
      );
    sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
    fxgv[i] = c_sub(  /* X-ray structure factor */
      zgv, c_mul( c_conv(131.2625 * sq_s / PI, 0.0), fg )
      );
    /* 8 * pi^2 * eps_0 * h^2 / m_e / e^2 = 131.2625 */
  }
  mb_fprintf(stdout, "100%%.\n");  mb_fflush(stdout);
  free_vector(fg_table);
  free_vector(g_table);
}


void xrd_fxgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex *fxgv
  )
{
  /*
   * X-ray structure factor fxgv
   */
  int		i, num_rf, ng, prog;
  int		ct_fg_table, g_table_size;
  HKL		g, *g_table;
  complex	fg, *fg_table;
  
  /* The constraints for low-order structure factors are ignored. */
  if(csd->num_lofg > 0){
    printf("$$ Note:\n");
    printf("$$  The constraints for low-order structure factors are to be ignored. \n");
  }
  
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  
  num_rf = num_exact_rf + num_bethe_rf;
  /* prog = 0; */
  for(i=0; i<num_rf; i++){

    progress_indicator(i, num_rf);
/*    exec_progress(&prog, i, num_rf); */
    g = reflex[i];
    
    if ( ( ng = search_g_table( ct_fg_table, reflex[i], g_table) ) >= 0 ){	/* found */
      fg = fg_table[ng];
    }
    else{												/* not found in table */
      fg = calc_Xray_Fg_ela(
        g, csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->xdyl, kvolts, bv
        );
      stock_fg_tbl_sym(
        g, fg, &ct_fg_table, g_table_size, g_table, fg_table,
        csd->num_sym_op, csd->sym_op
        );
    }
    fxgv[i] = fg;
  }
  
  mb_fprintf(stdout, "100%%.\n");
  free_vector(fg_table);
  free_vector(g_table);
}


void ncgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex *ncgv
  )
     /*
      * Nuclear charge structure factor ncgv
      */
{
  int		i, num_rf, ng, prog;
  int		ct_fg_table, g_table_size;
  HKL		g, *g_table;
  complex	*fg_table, zgv;
  
  if(csd->num_lofg > 0){
    printf("$$ Note:\n");
    printf("$$  The constraints for low-order structure factors are to be ignored. \n");
  }
  
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  
  num_rf = num_exact_rf + num_bethe_rf;
/*  prog = 0; */
  for(i=0; i<num_rf; i++){
    
    progress_indicator(i, num_rf);
/*    exec_progress(&prog, i, num_rf); */
    g = reflex[i];
    
    if ( ( ng = search_g_table( ct_fg_table, reflex[i], g_table) ) >= 0 ){	/* found */
      zgv = fg_table[ng];
    }
    else{ /* not found in the table */
      zgv = calc_z_factor(
        g, csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->xdyl, kvolts, bv, csd->lattice_type
        );
      stock_fg_tbl_sym(
        g, zgv, &ct_fg_table, g_table_size, g_table, fg_table,
        csd->num_sym_op, csd->sym_op
        );
    }
    ncgv[i] = zgv;
  }
  
  mb_fprintf(stdout, "100%%.\n");
  free_vector(fg_table);
  free_vector(g_table);
}


double calc_abs_K( CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv )
     /* K^2 = \chi^2 + U_0 */
{
  int nth;
  double abs_K, chi;
  complex f0, u0;
  HKL	g0;
  
  g0.h = 0;	g0.k = 0;	g0.l = 0;
  if(
    csd->num_lofg > 0
    && ( nth = find_n_g( csd->num_lofg, g0, csd->g_lofg ) ) != -1
    ){
    f0 = lofg_to_fg(
      csd->g_lofg[nth], csd->lof_type, &(csd->fg_lofg[nth]),
      csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm, csd->ftbl,
      csd->xdyl, kvolts, bv
      );
  }
  else{
    f0 = c_conv(
      calc_F_0_re(
        csd->num_sym_op, csd->n_atom, csd->atm,
        csd->xdyl, csd->ftbl, bv
        ),
      0.0
      );
  }
  u0 = fg_to_ug( f0, kvolts, bv );	/* U_0 */
  
  chi = 1.0 / wavelength(kvolts);
  abs_K = sqrt( sq(chi) + u0.re );
  
  /*	printf("$$ chi=%lf, abs_K=%lf [A^{-1}]\n", chi, abs_K); */
  return abs_K;
}


UVW downward(UVW inc)
     /* Return downward vector */
{
  UVW t;
  
  t.u = - inc.u; t.v = - inc.v; t.w = - inc.w;
  return t;
}


void incident_directions(
  int calc_type, int num_points,
  double H_begin, double H_end, double G_begin, double G_end,
  int *num_cal_points, int *h_pos, int *g_pos, R_HKL *k_xy,
  UVW incidence, R_HKL zolzbase1, CELL_PAR bv
  )
     /*
      * Size of $k_xy[]$ must be less than $(num_points)^{2}$.
      *  (calc_type 3)
      */
{
  int i, j, la, nm, ct = 0;
  R_HKL zolzbase2;
  double bh, bg;
  
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1 and has the same norm. */
  
  if (calc_type == 0){				/* 0 Dim. ( 1 point ) */
    bg = G_begin;
    bh = H_begin;
    k_xy[0].h = 0.5 * ( bh * zolzbase1.h + bg * zolzbase2.h );
    k_xy[0].k = 0.5 * ( bh * zolzbase1.k + bg * zolzbase2.k );
    k_xy[0].l = 0.5 * ( bh * zolzbase1.l + bg * zolzbase2.l );
    h_pos[0] = 0; g_pos[0] = 0;
    ct = 1;
  }
  else if (calc_type == 1){			/* 1 Dim. (Line prof.) */
    bg = G_begin;
    nm = num_points - 1;
    for(j=0; j<num_points; j++){
      bh = H_begin + ( H_end - H_begin ) * (double) j / (double) nm;
      k_xy[ct].h = 0.5 * ( bh * zolzbase1.h + bg * zolzbase2.h );
      k_xy[ct].k = 0.5 * ( bh * zolzbase1.k + bg * zolzbase2.k );
      k_xy[ct].l = 0.5 * ( bh * zolzbase1.l + bg * zolzbase2.l );
      h_pos[ct] = j - num_points / 2 - 1;
      g_pos[ct] = 0;
      ct++;
    }
  }
  else if (calc_type == 2){			/* 2 Dim. (Disk) */
    nm = num_points - 1;
    for(i=0; i<num_points; i++){
      bg = G_begin + ( G_end - G_begin ) * (double) i / (double) nm;
      la = 2 * (int)(
        sqrt(
          sq( (double)(num_points / 2) + 0.5 )
          - sq( (double)(- num_points / 2 + i) )
          )
        );
      for(j=0;j<la;j++){
        bh = H_begin + ( H_end - H_begin ) *
          (double) (num_points / 2 - la / 2 + j ) / (double) nm;
        k_xy[ct].h = 0.5 * ( bh * zolzbase1.h + bg * zolzbase2.h );
        k_xy[ct].k = 0.5 * ( bh * zolzbase1.k + bg * zolzbase2.k );
        k_xy[ct].l = 0.5 * ( bh * zolzbase1.l + bg * zolzbase2.l );
        h_pos[ct] = j - la / 2;
        g_pos[ct] = i - num_points / 2 - 1;
        ct++;
      }
    }
  }
  else if (calc_type == 3){			/* 2 Dim. (Square) */
    nm = num_points - 1;
    for(i=0; i<num_points; i++){
      bg = G_begin + ( G_end - G_begin ) * (double) i / (double) nm;
      for(j=0;j<num_points;j++){
        bh = H_begin + ( H_end - H_begin ) * (double) j / (double) nm;
        k_xy[ct].h = 0.5 * ( bh * zolzbase1.h + bg * zolzbase2.h );
        k_xy[ct].k = 0.5 * ( bh * zolzbase1.k + bg * zolzbase2.k );
        k_xy[ct].l = 0.5 * ( bh * zolzbase1.l + bg * zolzbase2.l );
        h_pos[ct] = j - num_points / 2 - 1; g_pos[ct] = i - num_points / 2 - 1;
        ct++;
      }
    }
  }
  else{
    mb_fprintf(stderr,"calc_type is invalid.\n");
    exit(EXIT_FAILURE);
  }
  
  *num_cal_points = ct;
}


double exct_err(
  double k_z, double abs_incidence, double abs_k,
  HKL g, R_HKL k_xy, UVW incidence, CELL_PAR bv
  )
     /*
      * excitation error $s_{g}$
      *   (k must be downwards, so $k_z < 0$. )
      *
      *  s_g = ( \chi^{2} - ( {\bf \chi} + {\bf g} )^{2} ) / ( 2 \chi )
      *      = ( - \chi_{z} g_{z} - {\bf \chi}_{proj} {\bf g}_{proj} - 0.5 g^{2} ) / \chi
      *      = ( - \chi_{z} g_{z} - {\bf \chi}_{proj} {\bf g} - 0.5 g^{2} ) / \chi
      */
{
  R_HKL t;
  double g_z;
  int n_lz;
  
  n_lz = g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
  g_z = (double) n_lz / abs_incidence;
  
  t.h = k_xy.h + 0.5 * (double) g.h;
  t.k = k_xy.k + 0.5 * (double) g.k;
  t.l = k_xy.l + 0.5 * (double) g.l;
  
  return ( - k_z * g_z - inner_product(t, hkl2rhkl(g), bv) ) / abs_k;
}


double tilting_factor(
  double k_z,  double abs_surf_norm, double cosfai_inc_surf,
  HKL g, R_HKL k_xy,  UVW surf_norm
  )
     /*
      * Beta_ge * chi
      *   (k must be downwards, so k_z < 0 )
      *
      *  $ ( {\bf \chi} + {\bf g} ) \cdot surf_norm $
      */
{
  R_HKL tg;
  
  tg.h = k_xy.h + (double) g.h;
  tg.k = k_xy.k + (double) g.k;
  tg.l = k_xy.l + (double) g.l;
  
  return  - k_z * cosfai_inc_surf
    + inner_product_uvw_hkl( downward(surf_norm), tg ) / abs_surf_norm;
}


void rho_and_beta_ge(
  int ne, int nb, double abs_incidence, double wave_length, 
  double abs_surf_norm, double cosfai_inc_surf,
  HKL *reflex, R_HKL k_xy, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double *rho, double *beta_ge
  )
{
  double abs_k, k_z;
  int i, nr;
  
  nr = ne + nb;
  abs_k = 1.0 / wave_length;
  k_z = - sqrt( sq(abs_k) - sq_abs(k_xy, bv) );	/* k: downwards. */
  for(i=0; i<nr; i++){
    rho[i] = wave_length
      * exct_err(
        k_z, abs_incidence, abs_k, reflex[i], k_xy, incidence, bv
        );
    beta_ge[i] = wave_length
      * tilting_factor(
        k_z, abs_surf_norm, cosfai_inc_surf,  reflex[i], k_xy, surf_norm
        );
  }
}


/* ----- */

double calc_K_z( double abs_K, R_HKL K_xy, CELL_PAR bv )
     /*
      * z-component of K (parallel to incidence); K is downwards.
      * incident vector is set to upward. (same as Zuo&Spence's prog.)
      */
{
  return - sqrt( sq(abs_K) - sq_abs(K_xy, bv) );
}


double calc_K_n(
  R_HKL K_xy, double K_z, double abs_surf_norm, UVW surf_norm, double cosfai_inc_surf
  )
     /*
      * normal component of K (parallel to surf_norm); K is downwards.
      * surface normal vector is set to upward. (same as Zuo&Spence's prog.)
      */
{
  return K_z * cosfai_inc_surf
    + inner_product_uvw_hkl( surf_norm, K_xy ) / abs_surf_norm;
}


double calc_g_z( HKL g, double abs_incidence, UVW incidence )
     /*
      * z-component of g (parallel to incidence)
      * surface normal vector is set to downward. (opposite to Zuo&Spence's prog.)
      */
{
  return inner_product_uvw_hkl( incidence, hkl2rhkl(g) ) / abs_incidence;
}


double calc_g_n( HKL g, double abs_surf_norm, UVW surf_norm )
     /* normal component of g (parallel to surf_norm) */
{
  return inner_product_uvw_hkl( surf_norm, hkl2rhkl(g) ) / abs_surf_norm;
}


double calc_excitation_error(
  R_HKL K_xy, double K_z, double abs_K, HKL g, double g_z, CELL_PAR bv
  )
     /*
      * Excitation error Sg
      *
      *  S_g = ( K^{2} - ( {\bf K} + {\bf g} )^{2} ) / ( 2K )
      *      = ( - K_{z} g_{z} - {\bf K}_{z} {\bf g}_{xy} - g^{2} / 2 ) / K
      *      = ( - K_{z} g_{z} - {\bf K}_{z} {\bf g} - g^{2} / 2 ) / K
      */
{
  R_HKL t;
  
  t.h = K_xy.h + (double) g.h / 2.0;
  t.k = K_xy.k + (double) g.k / 2.0;
  t.l = K_xy.l + (double) g.l / 2.0;
  
  return ( - K_z * g_z - inner_product(t, hkl2rhkl(g), bv) ) / abs_K;
}


void calc_sg_beta_g(
  R_HKL K_xy, double K_z, double K_n, double abs_K,
  double abs_incidence, UVW incidence, double abs_surf_norm, UVW surf_norm,
  int ne, int nb, HKL *reflex, CELL_PAR bv, double Sg[], double beta_g[]
  )
     /*
      * Sg and beta_g
      *
      *  S_g = ( K^{2} - ( {\bf K} + {\bf g} )^{2} ) / ( 2K )
      *	beta_g = 1 + g_n / K_n
      *
      *		K: downwards
      *		incidence, surf_norm: upwards
      *
      *		_z, _xy : incidence
      *		_n, _t  : surf_norm
      */
{
  double g_z, g_n;
  int i, nr;
  HKL	g;
  
  nr = ne + nb;
  
  for(i=0; i<nr; i++){
    g = reflex[i];
    g_z = calc_g_z( g, abs_incidence, incidence );
    Sg[i] = calc_excitation_error( K_xy, K_z, abs_K, g, g_z, bv );
    g_n = calc_g_n( g, abs_surf_norm, surf_norm );
    beta_g[i] = (1.0 + g_n / K_n);							/* <<<<< */
    if(dbg.sg_dump_flag == 1){
      mb_fprintf(
        stdout,
        "$$ Sg(%d %d %d) = %lf\n",
        reflex[i].h, reflex[i].k, reflex[i].l, Sg[i]
        );
    }
    if(dbg.betag_dump_flag == 1){
      mb_fprintf(
        stdout,
        "$$ beta_ge(%d %d %d) = %lf\n",
        reflex[i].h, reflex[i].k, reflex[i].l, beta_g[i]
        );
    }
    /*	printf("$$sg=%lf, beta_g=%lf\n", Sg[i], beta_g[i]);	*/
  }
}


/* ----- */

void pvg_to_vg( int ct_perturb, int nr, complex ***perturb_vg, complex **p_vg )
{
  int i,j;
  /* printf("ct_perturb=%d(one_pvg)\n",ct_perturb); */
  for(i=0; i<nr; i++){
    for(j=0; j<nr; j++){
      p_vg[i][j] = perturb_vg[ct_perturb][i][j];
    }
  }
}


void vg_to_pvg( int ct_perturb, int nr, complex **vg, complex ***perturb_vg )
{
  int i,j;
  /* printf("ct_perturb=%d(vg_to_pvg)\n",ct_perturb); */
  for(i=0; i<nr; i++){
    for(j=0; j<nr; j++){
      perturb_vg[ct_perturb][i][j] = vg[i][j];
    }
  }
}


void vg_select_wg_info(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex,
  complex **vg, HKL *reflex, double *rho,
  double wg_bethe_min, double wg_max, double accel_kvol
  )
{
  /*
   * information of vg selection
   *    ==> selinfo_nth_reflex[], n_sel_exact, n_sel_bethe
   */
  int i, ct_exact, ct_bethe;
  int *nth_reflex_exact, *nth_reflex_bethe;
  double abs_w_g, xi_g, relativistic_correction, wave_length;
  
  relativistic_correction = relativ_corr(accel_kvol);
  wave_length = wavelength(accel_kvol);
  
  nth_reflex_exact = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_exact);
  nth_reflex_bethe = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_bethe);
  
  ct_exact = 0; ct_bethe = 0; /* Counters for reflections exact and Bethe */
  nth_reflex_exact[0] = 0;	/* reflex_sel[0] must be 000 */
  ct_exact = 1;
  
  for(i = 1; i < n_exact + n_bethe; i++){
    xi_g = vg_to_xig( vg[i][0], relativistic_correction, wave_length );
    abs_w_g  = xi_g * fabs( rho[i] / wave_length );
    if ( abs_w_g < wg_bethe_min ){		/* Exact */
      nth_reflex_exact[ct_exact] = i;	ct_exact++;
    }
    else if ( abs_w_g < wg_max ){	/* Bethe */
      nth_reflex_bethe[ct_bethe] = i;	ct_bethe++;
    }
  }
  *n_sel_exact = ct_exact; *n_sel_bethe = ct_bethe;
  
  for(i = 0; i < *n_sel_exact; i++){				/* merge */
    selinfo_nth_reflex[i] = nth_reflex_exact[i];
  }
  for(i = 0; i < *n_sel_bethe; i++){
    selinfo_nth_reflex[i + *n_sel_exact] = nth_reflex_bethe[i];
  }
  
  free_vector(nth_reflex_exact);
  free_vector(nth_reflex_bethe);
}


void vg_select_sg_info(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex, HKL *reflex, double *rho,
  double sg_bethe_min, double sg_max, double accel_kvol
  )
     /*
      * information of vg selection
      *    ==> selinfo_nth_reflex[], n_sel_exact, n_sel_bethe
      */
{
  int i, ct_exact, ct_bethe;
  int *nth_reflex_exact, *nth_reflex_bethe;
  double abs_s_g, relativistic_correction, wave_length;
  
  relativistic_correction = relativ_corr(accel_kvol);
  wave_length = wavelength(accel_kvol);
  
  nth_reflex_exact = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_exact);
  nth_reflex_bethe = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_bethe);
  
  ct_exact = 0; ct_bethe = 0; /* Counters for reflections exact and Bethe */
  nth_reflex_exact[0] = 0;	/* reflex_sel[0] must be 000 */
  ct_exact = 1;
  
  for(i = 1; i < n_exact + n_bethe; i++){
    abs_s_g  = fabs( rho[i] / wave_length );
    if ( abs_s_g < sg_bethe_min ){		/* Exact */
      nth_reflex_exact[ct_exact] = i;	ct_exact++;
    }
    else if ( abs_s_g < sg_max ){	/* Bethe */
      nth_reflex_bethe[ct_bethe] = i;	ct_bethe++;
    }
  }
  *n_sel_exact = ct_exact; *n_sel_bethe = ct_bethe;
  
  for(i = 0; i < *n_sel_exact; i++){				/* merge */
    selinfo_nth_reflex[i] = nth_reflex_exact[i];
  }
  for(i = 0; i < *n_sel_bethe; i++){
    selinfo_nth_reflex[i + *n_sel_exact] = nth_reflex_bethe[i];
  }
  
  free_vector(nth_reflex_exact);
  free_vector(nth_reflex_bethe);
}


void vg_no_select_info(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex
  )
     /* Copy --- no effect */
{
  int i;
  
  *n_sel_exact = n_exact;	*n_sel_bethe = n_bethe;
  for(i = 0; i < n_exact + n_bethe; i++){				/* merge */
    selinfo_nth_reflex[i] = i;
  }
}


void vg_select_info(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex,
  complex **vg, HKL *reflex, double *rho,
  BEAM_SEARCH_PAR beam_search, double accel_kvol
  )
{
  if(epsilon.vg_select_flag != 0){
    if(
      beam_search.flag_bs == 1 || beam_search.flag_bs == 11
      ){			/* wg */
      vg_select_wg_info(
        n_exact, n_bethe, n_sel_exact, n_sel_bethe,
        selinfo_nth_reflex, vg, reflex,
        rho, beam_search.wgbmin,  beam_search.wgmax, accel_kvol
        );
    }
    else if(
      beam_search.flag_bs == 2 || beam_search.flag_bs == 12 ||
      beam_search.flag_bs == 3 || beam_search.flag_bs == 13
      ){/* sg */
      vg_select_sg_info(
        n_exact, n_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex, reflex,
        rho, beam_search.sgbmin,  beam_search.sgmax, accel_kvol
        );
    }
  }
  else {
    vg_no_select_info(
      n_exact, n_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex
      );
  }
}


void vg_select_trans(
  int n_exact, int n_bethe, int n_sel_exact, int n_sel_bethe,
  int *selinfo_nth_reflex,
  double *beta_ge, double *beta_ge_sel,
  int num_cal_g, int *nth_g, int *nth_g_sel
  )
{
  int i, j, nr_sel;
  
  nr_sel = n_sel_exact + n_sel_bethe;
  
  for(i = 0; i < nr_sel; i++){
    beta_ge_sel[i] = beta_ge[ selinfo_nth_reflex[i] ];
  }
  
  for(i=0; i<num_cal_g; i++){
    for(j=0; j<n_sel_exact; j++){
      if(nth_g[i] == selinfo_nth_reflex[j]){
        nth_g_sel[i] = j;	break;
      }
      else{ nth_g_sel[i] = -1; }
    }
  }
}


/* ----- */

void vg_select_wg_info_ug(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex,
  complex **vg, HKL *reflex, double *Sg,
  double wg_bethe_min, double wg_max, double accel_kvol
  )
     /*
      * information of vg selection
      *    ==> selinfo_nth_reflex[], n_sel_exact, n_sel_bethe
      */
{
  int i, ct_exact, ct_bethe;
  int *nth_reflex_exact, *nth_reflex_bethe;
  double abs_w_g, xi_g, relativistic_correction, wave_length;
  
  relativistic_correction = relativ_corr(accel_kvol);
  wave_length = wavelength(accel_kvol);
  
  nth_reflex_exact = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_exact);
  nth_reflex_bethe = new_vector(int, n_exact+n_bethe);
  chk_vector(nth_reflex_bethe);
  
  ct_exact = 0; ct_bethe = 0; /* Counters for reflections exact and Bethe */
  
  if( reflex[0].h == 0 && reflex[0].k == 0 && reflex[0].l == 0 ){
    nth_reflex_exact[0] = 0;	/* reflex_sel[0] must be 000 */
    ct_exact = 1;
  }
  else{
    mb_fprintf(
      stdout, "$$ (vg_select_wg_info_ug) reflex_sel[0] must be 000. Abort.\n"
      );
    exit(EXIT_FAILURE);
  }
  
  for(i = 1; i < n_exact + n_bethe; i++){
    xi_g = vg_to_xig( vg[i][0], relativistic_correction, wave_length );	/* <<<<< */
    abs_w_g  = xi_g * fabs( Sg[i] );
    if ( abs_w_g < wg_bethe_min ){		/* Exact */
      nth_reflex_exact[ct_exact] = i;	ct_exact++;
    }
    else if ( abs_w_g < wg_max ){	/* Bethe */
      nth_reflex_bethe[ct_bethe] = i;	ct_bethe++;
    }
  }
  *n_sel_exact = ct_exact; *n_sel_bethe = ct_bethe;
  
  for(i = 0; i < *n_sel_exact; i++){				/* merge */
    selinfo_nth_reflex[i] = nth_reflex_exact[i];
  }
  for(i = 0; i < *n_sel_bethe; i++){
    selinfo_nth_reflex[i + *n_sel_exact] = nth_reflex_bethe[i];
  }
  
  free_vector(nth_reflex_exact);
  free_vector(nth_reflex_bethe);
}


void vg_select_sg_info_ug(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex, HKL *reflex, double *Sg,
  double sg_bethe_min, double sg_max, double accel_kvol
  )
     /*
      * information of vg selection
      *    ==> selinfo_nth_reflex[], n_sel_exact, n_sel_bethe
      */
{
  int i, ct_exact, ct_bethe, numeb;
  int *nth_reflex_exact, *nth_reflex_bethe;
  double abs_s_g, relativistic_correction, wave_length;
  
  relativistic_correction = relativ_corr(accel_kvol);
  wave_length = wavelength(accel_kvol);
  
  numeb = n_exact + n_bethe;
  nth_reflex_exact = new_vector(int, numeb);	chk_vector(nth_reflex_exact);
  nth_reflex_bethe = new_vector(int, numeb);	chk_vector(nth_reflex_bethe);
  
  ct_exact = 0; ct_bethe = 0; /* Counters for reflections exact and Bethe */
  
  if( reflex[0].h == 0 && reflex[0].k == 0 && reflex[0].l == 0 ){
    nth_reflex_exact[0] = 0;	/* reflex_sel[0] must be 000 */
    ct_exact = 1;
  }
  else{
    mb_fprintf(
      stdout, "$$ (vg_select_sg_info_ug) reflex_sel[0] must be 000. Abort.\n"
      );
    exit(EXIT_FAILURE);
  }
  
  for(i = 1; i < numeb; i++){
    abs_s_g  = fabs( Sg[i] );
    if ( abs_s_g < sg_bethe_min ){		/* Exact */
      nth_reflex_exact[ct_exact] = i;	ct_exact++;
    }
    else if ( abs_s_g < sg_max ){	/* Bethe */
      nth_reflex_bethe[ct_bethe] = i;	ct_bethe++;
    }
  }
  
  *n_sel_exact = ct_exact; *n_sel_bethe = ct_bethe;
  
  for(i = 0; i < *n_sel_exact; i++){				/* merge */
    selinfo_nth_reflex[i] = nth_reflex_exact[i];
  }
  for(i = 0; i < *n_sel_bethe; i++){
    selinfo_nth_reflex[i + *n_sel_exact] = nth_reflex_bethe[i];
  }
  
  free_vector(nth_reflex_exact);
  free_vector(nth_reflex_bethe);
}


void vg_select_sg_info_ug_multi(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex, HKL *reflex, double *Sg,
  double *sge_max, double *sgb_max, 
  int lz_max, UVW incidence, double accel_kvol
  )
     /*
      * information of vg selection
      *    ==> selinfo_nth_reflex[], n_sel_exact, n_sel_bethe
      */
{
  int i, ct_exact, ct_bethe, numeb, nlz;
  int *nth_reflex_exact, *nth_reflex_bethe;
  double abs_s_g, relativistic_correction, wave_length;
  
  relativistic_correction = relativ_corr(accel_kvol);
  wave_length = wavelength(accel_kvol);
  
  numeb = n_exact + n_bethe;
  nth_reflex_exact = new_vector(int, numeb);	chk_vector(nth_reflex_exact);
  nth_reflex_bethe = new_vector(int, numeb);	chk_vector(nth_reflex_bethe);
  
  ct_exact = 0; ct_bethe = 0; /* Counters for reflections exact and Bethe */
  
  if( reflex[0].h == 0 && reflex[0].k == 0 && reflex[0].l == 0 ){
    nth_reflex_exact[0] = 0;	/* reflex_sel[0] must be 000 */
    ct_exact = 1;
  }
  else{
    mb_fprintf(
      stdout, "$$ (vg_select_sg_info_ug) reflex_sel[0] must be 000. Abort.\n"
      );
    exit(EXIT_FAILURE);
  }
  
  for(i = 1; i < numeb; i++){
    nlz = incidence.u * reflex[i].h
      + incidence.v * reflex[i].k
        + incidence.w * reflex[i].l;
    abs_s_g  = fabs( Sg[i] );
    if ( abs_s_g < sge_max[nlz] ){		/* Exact */
      nth_reflex_exact[ct_exact] = i;	ct_exact++;
    }
    else if ( abs_s_g < sgb_max[nlz] ){	/* Bethe */
      nth_reflex_bethe[ct_bethe] = i;	ct_bethe++;
    }
  }
  
  *n_sel_exact = ct_exact; *n_sel_bethe = ct_bethe;
  
  for(i = 0; i < *n_sel_exact; i++){				/* merge */
    selinfo_nth_reflex[i] = nth_reflex_exact[i];
  }
  for(i = 0; i < *n_sel_bethe; i++){
    selinfo_nth_reflex[i + *n_sel_exact] = nth_reflex_bethe[i];
  }
  
  free_vector(nth_reflex_exact);
  free_vector(nth_reflex_bethe);
}


void vg_no_select_info_ug(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex
  )
     /* Copy --- no effect */
{
  int i;
  
  *n_sel_exact = n_exact;	*n_sel_bethe = n_bethe;
  for(i = 0; i < n_exact + n_bethe; i++){				/* merge */
    selinfo_nth_reflex[i] = i;
  }
}


void vg_select_info_ug(
  int n_exact, int n_bethe, int *n_sel_exact, int *n_sel_bethe,
  int *selinfo_nth_reflex,
  complex **vg, HKL *reflex, double *Sg,
  BEAM_SEARCH_PAR beam_search, UVW incidence, double accel_kvol
  )
{
  if (epsilon.vg_select_flag != 0){
    if (
      beam_search.flag_bs == 1 || beam_search.flag_bs == 11
      ){	/* wg */
      vg_select_wg_info_ug(
        n_exact, n_bethe, n_sel_exact, n_sel_bethe,
        selinfo_nth_reflex, vg, reflex,
        Sg, beam_search.wgbmin,  beam_search.wgmax, accel_kvol
        );
    }
    else if (
      beam_search.flag_bs == 2 || beam_search.flag_bs == 12
      ){	/* sg */
      vg_select_sg_info_ug(
        n_exact, n_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex, reflex,
        Sg, beam_search.sgbmin,  beam_search.sgmax, accel_kvol
        );
    }
    else if (
      beam_search.flag_bs == 3 || beam_search.flag_bs == 13
      ){	/* sg multi */
      vg_select_sg_info_ug_multi(
        n_exact, n_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex, reflex,
        Sg, beam_search.sge_max,  beam_search.sgb_max, 
        beam_search.lzmax, incidence, accel_kvol
        );
    }
  }
  else {
    vg_no_select_info_ug(
      n_exact, n_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex
      );
  }
}


void vg_select_trans_ug(
  int n_exact, int n_bethe, int n_sel_exact, int n_sel_bethe, int *selinfo_nth_reflex,
  double *beta_g, double *beta_g_sel, int num_cal_g, int *nth_g, int *nth_g_sel
  )
{
  int i, j, nr_sel;
  
  nr_sel = n_sel_exact + n_sel_bethe;
  
  for(i = 0; i < nr_sel; i++){
    beta_g_sel[i] = beta_g[ selinfo_nth_reflex[i] ];
  }
  
  for(i=0; i<num_cal_g; i++){
    for(j=0; j<n_sel_exact; j++){
      if(nth_g[i] == selinfo_nth_reflex[j]){
        nth_g_sel[i] = j;	break;
      }
      else{ nth_g_sel[i] = -1; }
    }
  }
}


/* ----- */

void construct_matrix_selected(
  int num_sel_exact, int num_sel_bethe, int *selinfo_nth_reflex,
  double evdash, complex **vg, double *rho, double *beta_ge, complex **a_matrix
  )
{
  /*
   * Construction of the matrix to be diagonalized
   *   referring to vg_select_info.
   */
  int i, j, i_sel, j_sel, nr_sel;
  
  nr_sel = num_sel_exact + num_sel_bethe;
  
  for(i=0; i<nr_sel; i++){		/* off-diagonal term */
    i_sel = selinfo_nth_reflex[i];
    for(j=0; j<nr_sel; j++){ 
      j_sel = selinfo_nth_reflex[j];
      a_matrix[i][j]
        = c_mul(
          vg[i_sel][j_sel],
          c_conv(
            SCFACT / ( evdash * 2 * sqrt( beta_ge[i_sel] * beta_ge[j_sel] ) ),
            0.0
            )
          );
    }
  }
  for(i=0; i<nr_sel; i++){		/* diagonal term (V_0) */
    i_sel = selinfo_nth_reflex[i];
    a_matrix[i][i]
      = c_mul(
        c_add(
          c_div( vg[i_sel][i_sel], c_conv( evdash, 0.0 ) ),
          c_conv( 2.0 * rho[i_sel], 0.0 )
          ),
        c_conv( SCFACT / ( 2 * beta_ge[i_sel] ), 0.0 )
        );
  }
}


void construct_matrix(
  int ne, int nb, double evdash, 
  complex **vg, double *rho, double *beta_ge, complex **a_matrix
  )
{
  /*
   * Construction of the matrix to be diagonalized
   */
  int i, j, nr;
  
  nr = ne + nb;
  for(i=0; i<nr; i++){		/* off-diagonal term */
    for(j=0; j<nr; j++){ 
      a_matrix[i][j]
        = c_mul(
          vg[i][j],
          c_conv(
            SCFACT / ( evdash * 2 * sqrt( beta_ge[i] * beta_ge[j] ) ),
            0.0
            )
          );
    }
  }
  for(i=0; i<nr; i++){		/* diagonal term */
    a_matrix[i][i]
      = c_mul(
        c_add(
          c_div( vg[i][i], c_conv( evdash, 0.0 ) ),
          c_conv( 2.0 * rho[i], 0.0 )
          ),
        c_conv( SCFACT / ( 2 * beta_ge[i] ), 0.0 )
        );
  }
}


void construct_ug_mat_slct(
  int num_sel_exact, int num_sel_bethe, int *selinfo_nth_reflex,
  double accel_kvol, complex **vg, double *Sg, double *beta_g,
  complex **a_matrix
  )
{
  /*
   * Construction of the matrix to be diagonalized (Ug)
   *   referring to vg_select_info.
   *
   *  beta_g = 1 + g_n / K_n
   */
  int i, j, i_sel, j_sel, nr_sel;
  double rel_corr, abs_K, chi;
  complex u0;
  
  chi = 1.0 / wavelength(accel_kvol);
  rel_corr = relativ_corr(accel_kvol);
  u0 = vg_to_ug( vg[0][0], rel_corr );
  abs_K = sqrt( sq(chi) + u0.re );
  nr_sel = num_sel_exact + num_sel_bethe;
  
  for(i=0; i<nr_sel; i++){		/* off-diagonal term */
    i_sel = selinfo_nth_reflex[i];
    for(j=0; j<nr_sel; j++){ 
      j_sel = selinfo_nth_reflex[j];
      a_matrix[i][j]
        = c_div(
          vg_to_ug( vg[i_sel][j_sel], rel_corr ),
          c_conv( sqrt( beta_g[i_sel] * beta_g[j_sel] ), 0.0 )
          );
    }
  }
  for(i=0; i<nr_sel; i++){		/* diagonal term */
    i_sel = selinfo_nth_reflex[i];
    /* u0 = vg_to_ug( vg[i_sel][i_sel], rel_corr ); */
    a_matrix[i][i]
      = c_div(
        c_conv( 2.0 * abs_K * Sg[i_sel], u0.im ),	/* <<<<< */
        c_conv( beta_g[i_sel], 0.0 )
        );
  }
  /*	printf("$$ mat.\n");
	print_complexmat(nr_sel, a_matrix);*/
}


void g_bethe( int ne, int nb, complex **mat, complex **matb )
{
  /*
   * Generalized Bethe Approximation (GBA)
   *   Ichikawa & Hayakawa: JPSJ
   *
   * Input)   mat: matrix to be diagonalized (order ne+nb)
   *          ne : number of reflections treated as "exact".
   *          nb : number of reflections treated by "GBA".
   * Output)  matb: matrix modified by generalized Bethe Approx.
   *          (order ne)
   */
  int  i, j, err;
  complex  **wu, **wd, **wd2, **wc, **mcorr;

  if( nb == 0 ){  /* No Bethe reflections */
    c_matcopy2(ne, mat, matb);
    return;
  }

  wu  = new_matrix(complex, ne,nb);  chk_matrix(wu);
  wd  = new_matrix(complex, nb,ne);  chk_matrix(wd);
  wc  = new_matrix(complex, nb,nb);  chk_matrix(wc);
  wd2 = new_matrix(complex, nb,ne);  chk_matrix(wd2);
  mcorr = new_matrix(complex, ne,ne);  chk_matrix(mcorr);

  /*
   *           0 ----- (ne-1) (ne) -- (ne+nb-1)
   *         0 + ---------- +   + ----- +
   *           |            |   |       |
   *           |            |   |       |
   *           |            |   |  wu   |
   *           |            |   |       |
   *           |            |   |       |
   *           |            |   |       |
   *     (ne-1)+ ---------- +   + ----- +
   *
   *       (ne)+ ---------- +   + ----- +
   *           |            |   |       |
   *           |    wd      |   |  wc   |
   *           |            |   |       |
   * (ne+nb-1) + ---------- +   + ----- +
   *
   *   {matb} = {mat} - {wu} {wc}^-1 {wd}
   *    Approx.: eigenvalue $\gamma^{j}$ is set to be 0.
   */

  for(i=0;i<nb;i++){
    for(j=0;j<ne;j++){
      wd[i][j] = mat[ne+i][j];
    }
  }
  for(i=0;i<ne;i++){
    for(j=0;j<nb;j++){
      wu[i][j] = mat[i][ne+j];
    }
  }

  for(i=0;i<nb;i++){
    for(j=0;j<nb;j++){
      wc[i][j] = mat[ne+i][ne+j];
      /* Gamma^j correction may be done for this matrix.*/
    }
  }

  if ( (err = inv_c_mat(nb, epsilon.inv, wc)) != 0 ){
    mb_fprintf(stdout, "$$ error (g_bethe): inv_c_mat = %d.\n", err);
  }
  c_matmult( nb, ne, nb, wd2, wc, wd );
  c_matmult( ne, ne, nb, mcorr, wu, wd2);

  for(i=0;i<ne;i++){
    for(j=0;j<ne;j++){
      matb[i][j] = c_sub(mat[i][j], mcorr[i][j]);
    }
  }

  free_matrix(wu); free_matrix(wd); free_matrix(wc);
  free_matrix(wd2); free_matrix(mcorr);
}


void simple_bethe( int ne, int nb, complex **mat, complex **matb )
{
  /*
   * Simple Bethe Approximation
   *
   * Input)   mat: matrix to be diagonalized (order ne+nb)
   *          ne : number of reflections treated as "exact".
   *          nb : number of reflections treated by "Bethe approx.".
   * Output)  matb: matrix modified by simple Bethe Approx.
   *               (order ne)
   */
  int  i, j, k, maxn;
  complex **mcorr, ctmp;
  
  if( nb == 0 ){			/* No Bethe reflections */
    c_matcopy2(ne, mat, matb);
    return;
  }
  
  /*
   *           0 ----- (ne-1) (ne) -- (ne+nb-1)
   *         0 + ---------- +   + ----- +
   *           |            |   |       |
   *           |            |   |       |
   *           |   Exact    |   |       |
   *           |            |   |       |
   *           |            |   |       |
   *           |            |   |       |
   *     (ne-1)+ ---------- +   + ----- +
   *
   *       (ne)+ ---------- +   + ----- +
   *           |            |   |       |
   *           |            |   |       |
   *           |            |   |       |
   * (ne+nb-1) + ---------- +   + ----- +
   *
   */
  maxn = ne + nb;
  
  mcorr = new_matrix(complex, ne, ne);	chk_matrix(mcorr);
  for(i=0; i<ne; i++){
    for(j=0; j<ne; j++){
      ctmp = c_conv(0.0, 0.0);
      for(k=ne; k<maxn; k++){
        ctmp = c_add(
          ctmp, 
          c_div(
            c_mul( mat[i][k], mat[k][j] ),
            mat[k][k]
            )
          );
      }
      mcorr[i][j] = ctmp;
    }
  }
  for(i=0; i<ne; i++){
    for(j=0; j<ne; j++){
      matb[i][j] = c_sub(mat[i][j], mcorr[i][j]);
    }
  }
  free_matrix(mcorr);
  
  /* ***** Tests ***** */
  
  /* Other possibilities: I  ==> Slow! */
  /*
	for(i=0; i<ne; i++){
		for(j=0; j<ne; j++){
			ctmp = c_conv(0.0, 0.0);
			for(k=ne; k<maxn; k++){
				ctmp = c_add(
							ctmp, 
							c_div(
								c_mul( mat[i][k], mat[k][j] ),
								mat[k][k]
							)
						);
			}
			matb[i][j] = c_sub(mat[i][j], ctmp);
		}
	}
   */
  
  /* Other possibilities: II */
  /*
	for(i=0; i<ne; i++){
		for(j=0; j<ne; j++){
			matb[i][j] = mat[i][j];
		}
	}
    
	for(k=ne; k<maxn; k++){
		for(i=0; i<ne; i++){
			for(j=0; j<ne; j++){
			matb[i][j] = c_sub(
							matb[i][j], 
							c_div(
								c_mul( mat[i][k], mat[k][j] ),
								mat[k][k]
							)
						);
			}
		}
	}
   */
}



int eigmat_chk(
  int n,  double eps, complex **a, complex *eigen_val, complex **eigen_vec
  )
{
  /*
   * Check eigen values and eigen vectors
   */
  int i, j, err_flag;
  complex **ta, **ta2;
  
  mb_fprintf(stdout, "$$ eigen_c_mat checker\n");
  ta  = new_matrix(complex, n, n);	chk_matrix(ta);
  ta2 = new_matrix(complex, n, n);	chk_matrix(ta2);
  c_matcopy2(n, a, ta);
  c_matmult2(n, ta, eigen_vec);	/* left side */
  for(i=0;i<n;i++){				/* right side */
    for(j=0;j<n;j++){
      ta2[i][j] = c_mul( eigen_val[j], eigen_vec[i][j] );
    }
  }
  c_matsub2(n, ta, ta2);
  err_flag = 0;
  for(i=0;i<n;i++){
    for(j=0;j<n;j++){
      if(c_abs(ta[i][j]) > eps){
        mb_fprintf(
          stdout,
          "$$ (eigmat_chk) diff_element[%d][%d] = (%lg,%lg)\n",
          i, j, ta[i][j].re, ta[i][j].im
          );
        err_flag = 1;
      }
    }
  }
  free_matrix(ta);	free_matrix(ta2);
  if(err_flag == 0){
    mb_fprintf(stdout, "$$ eigen values and vectors are ok.\n");
    return 0;
  }
  else{
    mb_fprintf(stdout, "$$ error found in eigen_c_mat.");
    return 1;
  }
}


int upper_b_cond( int ne, double *beta_g, complex **eigen_vec, complex *bc)
{
  /*
   * Boundary condition at the upper surface
   *   success: return 0
   *   failed:  return 1
   * 0 must correspond to 000 beam.
   */
  int i, j, err_flag=0, err;
  double norm_sq;
  complex **tmp;
  
  tmp = new_matrix(complex, ne, ne);	chk_matrix(tmp);
  for(i=0; i<ne; i++){
    for(j=0;j<ne;j++){
      tmp[i][j] = c_div(eigen_vec[i][j], c_conv(sqrt( beta_g[i] ), 0.0));
    }
  }
  inverse( ne, tmp, bc, epsilon.inv, &err );
  if( err != 0 ){
    mb_fprintf(stdout, "$$ (upper_b_cond) inverse err = %d.\n", err);
  }
  
  /* chk */
  err_flag = 0;
  if(dbg.bc_chk_flag == 1 ){	/* debug level 1 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(
          stdout,
          "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
          i, bc[i].re, bc[i].im, c_abs(bc[i])
          );
        err_flag = 1;
      }
    }
  }
  else if(dbg.bc_chk_flag == 2){	/* debug level 2 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(stdout, "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
               i, bc[i].re, bc[i].im, c_abs(bc[i]) );
        norm_sq = 0.0;
        for(j=0; j<ne; j++){
          mb_fprintf(stdout, "$$ vec[%d][%d]=%lf %lf\n", j,i, eigen_vec[j][i]);
          norm_sq += sq(eigen_vec[j][i].re) + sq(eigen_vec[j][i].im);
        }
        printf("$$ norm^2 = %lf\n", norm_sq);
        err_flag = 1;
      }
    }
  }
  
  free_matrix(tmp);
  return err_flag;
}


int upper_b_cond_spcell(
  int ne, double *beta_g, complex *probe, complex **eigen_vec, complex *bc
  )
     /*
      * Boundary condition at the upper surface for the supercell mode.
      *   success: return 0
      *   failed:  return 1
      *	0 must correspond to 000 beam.
      */
{
  int i, j, err_flag=0;
  double norm_sq;
  complex **tmp, z;
  
  tmp = new_matrix(complex, ne, ne);	chk_matrix(tmp);
  
  /* boundary condition */
  
  for(i=0; i<ne; i++){
    for(j=0; j<ne; j++){
      tmp[i][j] = c_div(eigen_vec[i][j], c_conv(sqrt( beta_g[i] ), 0.0));
    }
  }
  
  inv_c_mat( ne, epsilon.inv, tmp );
  
  for(i=0; i<ne; i++){
    z = c_conv(0.0, 0.0);
    for(j=0; j<ne; j++){
      z = c_add( z, c_mul(tmp[i][j], probe[j]) );
    }
    bc[i] = z;
  }
  
  /* chk */
  err_flag = 0;
  if(dbg.bc_chk_flag == 1 ){	/* debug level 1 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(stdout, "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
               i, bc[i].re, bc[i].im, c_abs(bc[i]) );
        err_flag = 1;
      }
    }
  }
  else if(dbg.bc_chk_flag == 2){	/* debug level 2 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(stdout, "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
               i, bc[i].re, bc[i].im, c_abs(bc[i]) );
        norm_sq = 0.0;
        for(j=0; j<ne; j++){
          mb_fprintf(
            stdout,
            "$$ vec[%d][%d]=%lf %lf\n", j,i, eigen_vec[j][i]
            );
          norm_sq += sq(eigen_vec[j][i].re) + sq(eigen_vec[j][i].im);
        }
        mb_fprintf(stdout, "$$ norm^2 = %lf\n", norm_sq);
        err_flag = 1;
      }
    }
  }
  
  free_matrix(tmp);
  return err_flag;
}


int upper_b_cond_coherent(
  int ne, double *beta_g, complex initial_phase, complex **eigen_vec, complex *bc
  )
     /*
      * Boundary condition at the upper surface for coherent CBED mode.
      *   success: return 0
      *   failed:  return 1
      *	0 must correspond to 000 beam.
      */
{
  int i, j, err_flag=0, err;
  double norm_sq;
  complex **tmp;
  
  tmp = new_matrix(complex, ne, ne);	chk_matrix(tmp);
  
  /* boundary condition */
  
  for(i=0; i<ne; i++){
    for(j=0; j<ne; j++){
      tmp[i][j] = c_div(eigen_vec[i][j], c_conv(sqrt( beta_g[i] ), 0.0));
    }
  }
  
  inverse( ne, tmp, bc, epsilon.inv, &err );
  if( err != 0 ){
    mb_fprintf(stdout, "$$ (upper_b_cond) inverse err = %d.\n", err); }
  
  for(i=0; i<ne; i++){
    bc[i] = c_mul( bc[i], initial_phase );
  }
  
  /* chk */
  err_flag = 0;
  if(dbg.bc_chk_flag == 1 ){	/* debug level 1 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(stdout, "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
               i, bc[i].re, bc[i].im, c_abs(bc[i]) );
        err_flag = 1;
      }
    }
  }
  else if(dbg.bc_chk_flag == 2){	/* debug level 2 */
    for(i=0; i<ne; i++){
      if ( c_abs(bc[i]) > 1.0 ) {
        mb_fprintf(stdout, "$$ bc[%d] = %lf, %lf (c_abs = %lf)\n",
               i, bc[i].re, bc[i].im, c_abs(bc[i]) );
        norm_sq = 0.0;
        for(j=0; j<ne; j++){
          mb_fprintf(
            stdout, "$$ vec[%d][%d]=%lf %lf\n", j,i, eigen_vec[j][i]
            );
          norm_sq += sq(eigen_vec[j][i].re) + sq(eigen_vec[j][i].im);
        }
        mb_fprintf(stdout, "$$ norm^2 = %lf\n", norm_sq);
        err_flag = 1;
      }
    }
  }
  
  free_matrix(tmp);
  return err_flag;
}


#if 0
void pred_exec_time(long ind)
{
  /*
   * Prediction of calculation time of loop
   */
  static double c1,c2;
  double diff_t;
  
  if(ind == -1){ /* Initialize */
    mb_fprintf(stdout, " Estimating calc. time now...\n");
    mb_fflush(stdout);
    c1 = clock();
  }
  else if(ind == 0){		/* Termination */
    c2 = clock();
    mb_fprintf(stdout, "Elapsed time: %lf [min]\n",
           (double) (c2-c1)/ (double) CLOCKS_PER_SEC / 60.0 );
  }
  else{ /* Prediction (ind = number of the loop iteration) */
    c2 = clock();
    diff_t = (double) (c2-c1)/ (double) CLOCKS_PER_SEC;
    if( diff_t > 0.1 ){
      mb_fprintf(
        stdout, " Predicted calc. time = %.1lf [min]\n",
        diff_t * (double) ind / 60.0);
      mb_fprintf(stdout, " Calc.:0%%");
      mb_fflush(stdout);
    }
    else{
      mb_fprintf(stdout, " Prediction of calc. time has failed\n");
      mb_fprintf(stdout, "  because time for a cycle is too short.\n");
    }
  }
}
#endif


void limited_branch(
  double accuracy, int num_exact_rf, complex *boundary_cond,
  int *num_branch, int nth_branch[]
  )
{
  /*
   * Select branches according to accuracy
   *  0.0 < accuracy <= 1.0
   */
  int i, *index;
  double sum, *sq_norm_bc, all;
  
  for(i=0; i<num_exact_rf; i++){ nth_branch[i] = i; }	/* Init. */
  if( accuracy == 1.0 ) {		/* as it is */
    *num_branch = num_exact_rf;
    return;
  }
  
  sq_norm_bc = new_vector(double, num_exact_rf);  chk_vector(sq_norm_bc);
  index = new_vector(int, num_exact_rf);  chk_vector(index);
  all = 0.0;
  for(i=0; i<num_exact_rf; i++){
    sq_norm_bc[i] = sq(boundary_cond[i].re) + sq(boundary_cond[i].im);
    all += sq_norm_bc[i];
    /* printf("sq_norm_bc[%d] = %lf\n", i, sq_norm_bc[i]); */
  }
  /* ==> The value of all do not have to be 1.0 because of absorption effects */
  
  d_heapsort(num_exact_rf, sq_norm_bc, index);
  sum = 0.0;
  for(i=0; i<num_exact_rf; i++){
    sum += sq_norm_bc[i];
    nth_branch[i] = index[i];
    /* printf("$$ sum = %lf, nth_branch = %d\n", sum, nth_branch[i]); */
    if( sum / all >= accuracy ){	/* !!!!! */
      *num_branch = i + 1;
      break;
    }
  }
  
  free_vector(index);
  free_vector(sq_norm_bc);
}


double psi_to_intensity(
  int nth_g, double thickness, int num_exact_rf, double wave_length,
  double *beta_ge, complex *eigen_value, complex **eigen_vec,
  complex *boundary_cond
  )
     /*
      * Calculate intensities of reflections from Bloch wave functions
      */
{
  int j;
  complex psi;
  
  if( nth_g < 0 || nth_g >= num_exact_rf ) {
    /* printf("$$ nth_g is out of range. intensity is set to zero.\n"); */
    /* printf("."); fflush(stdout); */
    return 0.0;
  }
  
  psi = c_conv(0.0,0.0);
  for(j=0; j<num_exact_rf; j++){
    psi = c_add(
      psi,
      c_mul(
        c_mul(
          boundary_cond[j],
          c_div(
            eigen_vec[nth_g][j],
            c_conv( sqrt(beta_ge[nth_g]), 0.0 )		/* <<<<< */
            )
          ),
        c_exp(
          c_mul(
            eigen_value[j],
            c_conv(
              0.0,
              2.0 * PI * thickness / wave_length / SCFACT
              )
            )
          )
        )
      );
  }
  return sq(psi.re) + sq(psi.im);
}


double psi_to_intensity_limited(
  int nth_g, double thickness, int num_exact, 
  int num_branch, int *nth_branch, double wave_length,
  double *beta_ge, complex *eigen_value, complex **eigen_vec,
  complex *boundary_cond
  )
     /*
      * Calculate intensities of reflections
      *  from limited Bloch wave functions.
      * num_branch: total number of branches to be included
      * nth_branch[0-(num_branch-1)]: number of each branch
      */
{
  int j, k;
  complex psi;
  /*	complex tmp_z;*/
  
  if( nth_g < 0 || nth_g >= num_exact ) {
    /* printf("$$ nth_g is out of range. intensity is set to zero.\n"); */
    /* printf("."); fflush(stdout); */
    return 0.0;
  }
  /*	printf("\n$$ ***** nth_g = %d, thickness = %lf\n", nth_g, thickness);*/
  
  psi = c_conv(0.0,0.0);
  for(j=0; j<num_branch; j++){
    k = nth_branch[j];
    psi = c_add(
      psi,
      c_mul(
        c_mul(
          boundary_cond[k],
          c_div(
            eigen_vec[nth_g][k],
            c_conv( sqrt(beta_ge[nth_g]), 0.0 )		/* <<<<< */
            )
          ),
        c_exp(
          c_mul(
            eigen_value[k],
            c_conv(
              0.0,
              2.0 * PI * thickness / wave_length / SCFACT
              )
            )
          )
        )
      );
    
    /*		printf("$$ branch=%d\n",k);
		printf("$$  eps^j = (%lf, %lf)\n", boundary_cond[k].re, boundary_cond[k].im);
		printf("$$  Cg^j  = (%lf, %lf)\n", eigen_vec[nth_g][k].re, eigen_vec[nth_g][k].im);
			tmp_z = c_mul(boundary_cond[k], eigen_vec[nth_g][k]);
		printf("$$  epsCg = (%lf, %lf)\n", tmp_z.re, tmp_z.im);
		printf("$$  gamma^j = (%lf, %lf)\n", eigen_value[k].re, eigen_value[k].im);
			tmp_z = c_exp(
							c_mul(
								eigen_value[k],
								c_conv(
								    0.0,
									2.0 * PI * thickness / wave_length / SCFACT
								)
							)
						);
		printf("$$  exp-term = (%lf, %lf)\n", tmp_z.re, tmp_z.im);
		printf("$$  tmp-psi = (%lf, %lf)\n", psi.re, psi.im);
     */
  }
  return sq(psi.re) + sq(psi.im);
}


double psi_to_intensity_limited_ug(
  int nth_g, double thickness, int num_exact,
  int num_branch, int *nth_branch, double K_n,
  double *beta_g, complex *eigen_value, complex **eigen_vec,
  complex *boundary_cond
  )
{
  /*
   * Calculate intensities of reflections
   *  from limited Bloch wave functions.
   * num_branch: total number of branches to be included
   * nth_branch[0-(num_branch-1)]: number of each branch
   *
   * sign convention:
   * $ \phi = \sum \eps^j * C_g^j * \exp( 2 \pi i {\bf k}_g^j {\bf r} )
   *
   * Caution:
   *  minus sign in exp term is due to the direction of n (upwards).
   *  because z = -t at thickness t.
   */
  int j, k;
  complex psi;
  /* complex tmp_z;*/

  if( nth_g < 0 || nth_g >= num_exact ) {
    /* printf("$$ nth_g is out of range. intensity is set to zero.\n"); */
    /* printf("."); fflush(stdout); */
    return 0.0;
  }
  /*printf("\n$$ ***** nth_g = %d, thickness = %lf\n", nth_g, thickness);*/

  psi = c_conv(0.0,0.0);
  for(j=0; j<num_branch; j++){
    k = nth_branch[j];
    psi = c_add(
      psi,
      c_mul(
        c_mul(
          boundary_cond[k],
          c_div(
            eigen_vec[nth_g][k],
            c_conv( sqrt(beta_g[nth_g]), 0.0 )
            )
          ),
        c_exp(
          c_mul(
            eigen_value[k], c_conv( 0.0, - PI * thickness / K_n )
            /*  ^^^	 */
            )
          )
        )
      );
    
#if 0
    mb_fprintf(stdout, "$$ branch=%d\n",k);
    mb_fprintf(
      stdout, "$$  eps^j = (%lf, %lf)\n",
      boundary_cond[k].re, boundary_cond[k].im
      );
    mb_fprintf(
      stdout, "$$  Cg^j  = (%lf, %lf)\n",
      eigen_vec[nth_g][k].re / sqrt(beta_g[nth_g]),
      eigen_vec[nth_g][k].im / sqrt(beta_g[nth_g])
      );
    tmp_z = c_mul(
      boundary_cond[k],
      c_div(
        eigen_vec[nth_g][k],
        c_conv( sqrt(beta_g[nth_g]), 0.0 )
        )
      );
    mb_fprintf(stdout, "$$  epsCg = (%lf, %lf)\n", tmp_z.re, tmp_z.im);
    mb_fprintf(
      stdout,
      "$$  gamma^j = (%lf, %lf)\n", eigen_value[k].re, eigen_value[k].im
      );
    tmp_z = c_exp(
      c_mul(
        eigen_value[k],
        c_conv( 0.0, PI * thickness / K_n )
        )
      );
    mb_fprintf(stdout, "$$  exp-term = (%lf, %lf)\n", tmp_z.re, tmp_z.im);
    mb_fprintf(stdout, "$$  tmp-psi = (%lf, %lf)\n", psi.re, psi.im);
#endif
    
  }
#if 0
/* Only for test mode */
  mb_fprintf(
    stdout, " thickness, psi.re, psi.im, int: %lf %lf %lf %lf\n",
    thickness, psi.re, psi.im, sq(psi.re) + sq(psi.im)
    );
#endif
  return sq(psi.re) + sq(psi.im);
}


complex psi_limited_ug(
  int nth_g, double thickness, int num_exact,
  int num_branch, int *nth_branch, double K_n,
  double *beta_g, complex *eigen_value, complex **eigen_vec,
  complex *boundary_cond
  )
     /*
      * Calculate a wave funcion of reflections
      *  from limited Bloch wave functions.
      * num_branch: total number of branches to be included
      * nth_branch[0-(num_branch-1)]: number of each branch
      *
      * sign convention:
      *	$ \phi = \sum \eps^j * C_g^j * \exp( 2 \pi i {\bf k}_g^j {\bf r} )
      *
      * Caution:
      *	minus sign in exp term is due to the direction of n (upwards).
      *	because z = -t at thickness t.
      */
{
  int j, k;
  complex psi;
  /*	complex tmp_z;*/
  
  if( nth_g < 0 || nth_g >= num_exact ) {
    /* printf("$$ nth_g is out of range. intensity is set to zero.\n"); */
    /* printf("."); fflush(stdout); */
    return c_conv(0.0, 0.0);
  }
  /*printf("\n$$ ***** nth_g = %d, thickness = %lf\n", nth_g, thickness);*/
  
  psi = c_conv(0.0,0.0);
  for(j=0; j<num_branch; j++){
    k = nth_branch[j];
    psi = c_add(
      psi,
      c_mul(
        c_mul(
          boundary_cond[k],
          c_div(
            eigen_vec[nth_g][k],
            c_conv( sqrt(beta_g[nth_g]), 0.0 )
            )
          ),
        c_exp(
          c_mul(
            eigen_value[k], c_conv( 0.0, - PI * thickness / K_n )
            /*  ^^^	 */
            )
          )
        )
      );
    
  }
  return psi;
}


void pre_calc_intensity(
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double *abs_incidence, double *abs_surf_norm, double *cosfai_inc_surf,
  double *wave_length, double *ev_dash
  )
{
  *wave_length = wavelength(accel_kvol);
  *ev_dash = evdash(accel_kvol);
  *abs_incidence = abs_uvw(incidence, bv);  /* incidence is upwards */
  *abs_surf_norm = abs_uvw(surf_norm, bv);  /* surf_norm is upwards */
  *cosfai_inc_surf
    = inner_product_uvw( downward(incidence), downward(surf_norm), bv )
      / ( *abs_incidence * *abs_surf_norm );
}


void alloc_work_cal_int(
  int nr, int num_cal_g, int num_perturb, WORK_CAL_INT *wci
  )
{
  wci->Sg      = new_vector( double, nr );
  chk_vector(wci->Sg);
  wci->beta_g  = new_vector( double, nr );
  chk_vector(wci->beta_g);
  wci->beta_g_sel  = new_vector( double, nr );
  chk_vector(wci->beta_g_sel);
  wci->a_matrix = new_matrix( complex, nr, nr );
  chk_matrix(wci->a_matrix);
  wci->b_matrix = new_matrix( complex, nr, nr );
  chk_matrix(wci->b_matrix);
  wci->eigen_vec = new_matrix( complex, nr, nr );
  chk_matrix(wci->eigen_vec);
  wci->eigen_value = new_vector( complex, nr );
  chk_vector(wci->eigen_value);
  wci->boundary_cond = new_vector( complex, nr );
  chk_vector(wci->boundary_cond);
  wci->nth_branch = new_vector(int, nr);
  chk_vector(wci->nth_branch);
  wci->nth_g_sel = new_vector(int, num_cal_g);
  chk_vector(wci->nth_g_sel);
  wci->selinfo_nth_reflex = new_vector(int, nr);
  chk_vector(wci->selinfo_nth_reflex);

  wci->num_perturb = num_perturb;
  if( num_perturb > 0 ){
    wci->eigen_vec_inv = new_matrix(complex, nr, nr);
    chk_matrix(wci->eigen_vec_inv);
    wci->p_vg = new_matrix(complex, nr, nr);
    chk_matrix(wci->p_vg);
    wci->pa_mat = new_matrix(complex, nr, nr);
    chk_matrix(wci->pa_mat);
    wci->pb_mat = new_matrix(complex, nr, nr);
    chk_matrix(wci->pb_mat);
    wci->p_eigen_vec = new_matrix(complex, nr, nr);
    chk_matrix(wci->p_eigen_vec);
    wci->p_eigen_val = new_vector(complex, nr);
    chk_vector(wci->p_eigen_val);
    wci->p_boundary_cond = new_vector(complex, nr);
    chk_vector(wci->p_boundary_cond);
  }
}


void free_work_cal_int( WORK_CAL_INT *wci )
{
  free_vector(wci->Sg);
  free_vector(wci->beta_g);
  free_vector(wci->beta_g_sel);
  free_matrix(wci->a_matrix);
  free_matrix(wci->b_matrix);
  free_matrix(wci->eigen_vec);
  free_vector(wci->eigen_value);
  free_vector(wci->boundary_cond);
  free_vector(wci->nth_branch);
  free_vector(wci->nth_g_sel);
  free_vector(wci->selinfo_nth_reflex);
  if( wci->num_perturb > 0 ){
    free_matrix(wci->eigen_vec_inv);
    free_matrix(wci->p_vg);
    free_matrix(wci->pa_mat);
    free_matrix(wci->pb_mat);
    free_matrix(wci->p_eigen_vec);
    free_vector(wci->p_eigen_val);
    free_vector(wci->p_boundary_cond);
  }
}


#if 0
void calc_intensity_sel(
  R_HKL k_xy, BEAM_SEARCH_PAR beam_search,
  int ne, int nb, HKL *reflex, complex **vg,
  int num_cal_g, int nth_g[], int num_cal_thick, double thickness[],
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
  double wave_length, double ev_dash, double **intensity
  )
{
  /*
   * Calculation of diffracted intensities and perturbed intensities
   *   (for a incident direction)
   */
  int it, ig, nr, num_branch, *nth_branch;
  int n_sel_exact, n_sel_bethe, *selinfo_nth_reflex, *nth_g_sel;
  double *rho, *beta_ge, *beta_ge_sel;
  complex **a_matrix, **b_matrix, **eigen_vec, *eigen_value, *boundary_cond;
  
  nr = ne + nb;
  rho      = new_vector( double, nr );		chk_vector(rho);
  beta_ge  = new_vector( double, nr );		chk_vector(beta_ge);
  beta_ge_sel  = new_vector( double, nr );	chk_vector(beta_ge_sel);
  a_matrix = new_matrix( complex, nr, nr );	chk_matrix(a_matrix);
  b_matrix = new_matrix( complex, nr, nr );	chk_matrix(b_matrix);
  eigen_vec = new_matrix( complex, nr, nr );	chk_matrix(eigen_vec);
  eigen_value = new_vector( complex, nr );	chk_vector(eigen_value);
  boundary_cond = new_vector( complex, nr );	chk_vector(boundary_cond);
  nth_branch = new_vector(int, nr);			chk_vector(nth_branch);
  nth_g_sel = new_vector(int, num_cal_g);		chk_vector(nth_g_sel);
  selinfo_nth_reflex = new_vector(int, nr);	chk_vector(selinfo_nth_reflex);
  
  /* calc. of rho and beta_ge */
  dbg_mon("$$ rho_and_beta_ge\n");
  rho_and_beta_ge(
    ne, nb, abs_incidence, wave_length, abs_surf_norm, cosfai_inc_surf,
    reflex, k_xy, incidence, surf_norm, bv, rho, beta_ge
    );
  
  /* vg selection */
  dbg_mon("$$ vg_select\n");
  vg_select_info(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, rho, beam_search, accel_kvol
    );
  vg_select_trans(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_ge, beta_ge_sel, num_cal_g, nth_g, nth_g_sel
    );
  
  /* debug */
  if(dbg.vg_sel_chk_flag == 1){
    dbg_sel_vg(k_xy, ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex);
    free_vector(rho);		free_vector(beta_ge);	free_vector(beta_ge_sel);
    free_matrix(a_matrix);	free_matrix(b_matrix);
    free_matrix(eigen_vec);	free_vector(eigen_value);
    free_vector(boundary_cond);
    free_vector(nth_branch);	free_vector(nth_g_sel);
    free_vector(selinfo_nth_reflex);
    return;
  }
  
  /* matrix */
  dbg_mon("$$ construct_matrix\n");
  construct_matrix_selected(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    ev_dash, vg, rho, beta_ge, a_matrix
    );
  
  /* Generalized Bethe approximation */
  dbg_mon("$$ generalized Bethe approximation\n");
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  
  /* Eigen value problem (QR method) */
  dbg_mon("$$ eigen values and eigen vectors\n");		
  eispack_cg( n_sel_exact, epsilon.eigen, b_matrix, eigen_value, eigen_vec );
#if 0
  eigen_c_mat( ne, epsilon.eigen, ind = 1, b_matrix, eigen_value, eigen_vec )
#endif
    if( dbg.eigval_dump_flag == 1 ){
      mb_fprintf(dbg.fp,"eigen values:\n");
      fprint_complexvec(dbg.fp, n_sel_exact, eigen_value);
    }
  if( dbg.eigvec_dump_flag == 1 ){
    /*		fprintf(dbg.fp,"eigen vectors: ik=%d\n",ik); */
    mb_fprintf(dbg.fp,"eigen vectors:\n");
    fprint_complexmat(dbg.fp, n_sel_exact, eigen_vec);
  }
  
  /* Boundary condition */
  dbg_mon("$$ upper_b_cond\n");
  upper_b_cond( n_sel_exact, beta_ge_sel, eigen_vec, boundary_cond );
  
  /* Branch selection (for reduction of calc. speed) */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  
  /* Calc. intensity */
  dbg_mon("$$ calc_intensity\n");
  for(ig=0; ig<num_cal_g; ig++){
    for(it=0; it<num_cal_thick; it++){
      intensity[ig][it] =
        psi_to_intensity_limited(
          nth_g_sel[ig], thickness[it], n_sel_exact,
          num_branch, nth_branch, wave_length,
          beta_ge_sel, eigen_value, eigen_vec, boundary_cond
          );
    }
  }
  
  free_vector(rho);		free_vector(beta_ge);	free_vector(beta_ge_sel);
  free_matrix(a_matrix);	free_matrix(b_matrix);
  free_matrix(eigen_vec);	free_vector(eigen_value);
  free_vector(boundary_cond);
  free_vector(nth_branch);	free_vector(nth_g_sel);
  free_vector(selinfo_nth_reflex);
}
#endif


#if 0
void calc_perturb(
  int ne, int nb, HKL *reflex, int flag_bs,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int num_cal_thick, double thickness[],
  double wave_length, double ev_dash,
  int num_branch, int *nth_branch,
  int n_sel_exact, int n_sel_bethe, int *selinfo_nth_reflex, int *nth_g_sel,
  double *rho, double *beta_ge, double *beta_ge_sel,
  complex **matrix_unperturb, complex **eigen_vec, complex *eigen_value,
  double ***p_intensity, WORK_CAL_INT *wci
  )
{
  /*
   * Calculation of perturbed intensities
   *   (for a incident direction)
   *  with vg selection
   *
   * *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
   */
  int it, ig, ip, nr, nr_sel, err;
  complex **p_vg, **pa_mat, **pb_mat, *p_eigen_val, **p_eigen_vec, *p_boundary_cond;
  complex **eigen_vec_inv;
  
  if(num_perturb == 0){ return; }	/* No calculation */
  
  nr = ne + nb;
  nr_sel = n_sel_exact + n_sel_bethe;
  
  /* ----- copy pointers of work area wci ----- */
  eigen_vec_inv = wci->eigen_vec_inv;
  p_vg = wci->p_vg;	pa_mat = wci->pa_mat;	pb_mat = wci->pb_mat;
  p_eigen_vec = wci->p_eigen_vec;	p_eigen_val = wci->p_eigen_val;
  p_boundary_cond = wci->p_boundary_cond;
  
  /* ----- Inverse matrix of the unperturbed eigen vector ----- */
  dbg_mon("$$ eigen_vec_inv\n");
  c_matcopy2( n_sel_exact, eigen_vec, eigen_vec_inv);
  /* eigen_vec --> eigen_vec_inv */
  if ( (err = inv_c_mat( n_sel_exact, epsilon.inv, eigen_vec_inv ) ) != 0 ){
    mb_fprintf(stdout, "$$ error: inv_c_mat = %d.\n", err);
  }
  
  for(ip=0; ip<num_perturb; ip++){
    
    /* ----- Transfer perturb_vgs to p_vg ----- */
    dbg_mon("$$ transfer perturbed vg\n");
    pvg_to_vg(ip, nr, perturb_vgs, p_vg);
    
    /* Perturb: Vg selection  ==> make use of unperturbed selinfo */
    
    /* ----- Perturb: Matrix ----- */	dbg_mon("$$ p-construct_matrix\n");
    construct_matrix_selected(
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
      ev_dash, p_vg, rho, beta_ge, pa_mat
      );
    /*
		if( dbg.a_mat_dump_flag	== 1 ){
			fprintf(dbg.efp,"a_matrix: (perturb)\n");
			fprint_complexmat(dbg.efp, ne, pa_mat);
		}
     */
    
    /* ----- Perturb: Generalized Bethe approximation ----- */	dbg_mon("$$ p-g-bethe\n");
    if(flag_bs < 10){
      g_bethe( n_sel_exact, n_sel_bethe, pa_mat, pb_mat );
    }
    else if(flag_bs >= 10){
      simple_bethe( n_sel_exact, n_sel_bethe, pa_mat, pb_mat );
    }
    /*
		if( dbg.b_mat_dump_flag	== 1 ){
			fprintf(dbg.efp,"bethe_matrix: (perturb)\n");
			fprint_complexmat(dbg.efp, ne, pb_mat);
		}*/
    
    /* ----- Perturbation ----- */	dbg_mon("$$ perturbation2\n");
    perturbation2(												/* <<<<< */
                                                                eigen_value, eigen_vec, eigen_vec_inv, n_sel_exact, matrix_unperturb, pb_mat,
                                                                epsilon.degeneracy, epsilon.eigen, p_eigen_val, p_eigen_vec
                                                                );
    /*
			perturbation(
				eigen_val, eigen_vec, eigen_vec_inv, ne, bethe_mat, pb_mat,
				epsilon.degeneracy, epsilon.eigen, p_eigen_val, p_eigen_vec
			);
			eigen_val_perturbation(
				eigen_val, eigen_vec, eigen_vec_inv,
				ne, bethe_mat, pb_mat, p_eigen_val
			);
     */
    /*
		if( dbg.eigval_dump_flag == 1 ){
			fprintf(dbg.efp, "parameter %d: perturbed eigen values:\n", ip);
			fprint_complexvec(dbg.efp, n_sel_exact, p_eigen_val);
		}
		if( dbg.eigvec_dump_flag == 1 ){
			fprintf(dbg.efp, "parameter %d: perturbed eigen vectors:\n", ip);
			fprint_complexmat(dbg.efp, n_sel_exact, p_eigen_vec);
		}*/
    
    /* ----- Perturb: boundary condition ----- */	dbg_mon("$$ p-upper_b_cond\n");
    upper_b_cond( n_sel_exact, beta_ge_sel, p_eigen_vec, p_boundary_cond );
    
    /* ----- Perturb: calc. of perturbed intensities ----- */ dbg_mon("$$ p-intensities\n");
    for(ig=0; ig<num_cal_g; ig++){
      for(it=0; it<num_cal_thick; it++){
        p_intensity[ip][ig][it]
          = psi_to_intensity_limited(
            nth_g_sel[ig], thickness[it], n_sel_exact,
            num_branch, nth_branch, wave_length,
            beta_ge_sel, p_eigen_val, p_eigen_vec, p_boundary_cond
            );						   /* ^^^^^^^^^^^  ^^^^^^^^^^^^^^^ */
      }
    }
  }
  
  dbg_mon("$$perturbation end\n");
}
#endif


#if 0
void calc_intensity_with_perturb(
  R_HKL k_xy, BEAM_SEARCH_PAR beam_search,
  int ne, int nb, HKL *reflex, complex **vg,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
  double wave_length, double ev_dash,
  double **intensity,	double ***p_intensity, WORK_CAL_INT *wci
  )
{
  /*
   * Calculation of diffracted intensities and perturbed intensities
   *   (for a incident direction)
   *  with vg selection
   *
   * *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
   */
  int it, ig, nr, num_branch, *nth_branch;
  int n_sel_exact, n_sel_bethe, *selinfo_nth_reflex, *nth_g_sel;
  double *rho, *beta_ge, *beta_ge_sel;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond;
  
  nr = ne + nb;
  
  /* ----- copy pointers of work area wci ----- */
  rho      = wci->Sg;	beta_ge  = wci->beta_g;	beta_ge_sel  = wci->beta_g_sel;
  a_matrix = wci->a_matrix;	b_matrix = wci->b_matrix;
  eigen_vec = wci->eigen_vec;	eigen_value = wci->eigen_value;
  boundary_cond = wci->boundary_cond;
  nth_branch = wci->nth_branch;	nth_g_sel = wci->nth_g_sel;
  selinfo_nth_reflex = wci->selinfo_nth_reflex;
  
  /* ----- calc. of rho and beta_ge ----- */	dbg_mon("$$ rho_and_beta_ge\n");
  rho_and_beta_ge(
    ne, nb, abs_incidence, wave_length, abs_surf_norm, cosfai_inc_surf,
    reflex, k_xy, incidence, surf_norm, bv, rho, beta_ge
    );
  
  /* ----- vg selection ----- */	dbg_mon("$$ vg_select\n");
  vg_select_info(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, rho, beam_search, accel_kvol
    );
  vg_select_trans(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_ge, beta_ge_sel, num_cal_g, nth_g, nth_g_sel
    );
  
  /* ----- matrix ----- */	dbg_mon("$$ construct_matrix\n");
  construct_matrix_selected(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    ev_dash, vg, rho, beta_ge, a_matrix
    );
  
  /* ----- Generalized Bethe approximation ----- */	dbg_mon("$$ GBA\n");
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  
  /* ----- Eigen value problem (QR method) ----- */	dbg_mon("$$ eigen values and vectors\n");
  c_matcopy2( n_sel_exact, b_matrix, a_matrix);	/* b_matrix --> a_matrix */	
  eispack_cg( n_sel_exact, epsilon.eigen, a_matrix, eigen_value, eigen_vec );
  /* a_matrix is broken by eispack_cg. */
#if 0
  if ((err = eigen_c_mat( ne, epsilon.eigen, ind = 1, b_matrix, eigen_value, eigen_vec )) != 0){
    mb_fprintf(stdout, "$$ eigen_c_mat = %d\n",err);
  }
#endif
  /*if( dbg.eigval_dump_flag == 1 ){
		fprintf(dbg.fp,"eigen values:\n");
		fprint_complexvec(dbg.fp, n_sel_exact, eigen_value);
	}
	if( dbg.eigvec_dump_flag == 1 ){
		fprintf(dbg.fp,"eigen vectors:\n");
		fprint_complexmat(dbg.fp, n_sel_exact, eigen_vec);
	}*/
  
  /* ----- Boundary condition ----- */	dbg_mon("$$ upper_b_cond\n");
  upper_b_cond( n_sel_exact, beta_ge_sel, eigen_vec, boundary_cond );
  
  /* ----- Branch selection (for reduction of calc. speed) ----- */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  
  /* ----- Calc. intensity ----- */		dbg_mon("$$ calc_intensity\n");
  for(ig=0; ig<num_cal_g; ig++){
    for(it=0; it<num_cal_thick; it++){
      intensity[ig][it]
        = psi_to_intensity_limited(
          nth_g_sel[ig], thickness[it], n_sel_exact,
          num_branch, nth_branch, wave_length,
          beta_ge_sel, eigen_value, eigen_vec, boundary_cond
          );
    }
  }
  
  /*
   * Perturbation
   *   using perturbed potential perturb_vgs[ip][i][j],
   *   eigen values and eigen vectors
   */
  
  if(num_perturb > 0){
    calc_perturb(
      ne, nb, reflex, beam_search.flag_bs, num_perturb, perturb_vgs,
      num_cal_g, num_cal_thick, thickness,
      wave_length, ev_dash, num_branch, nth_branch,
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex, nth_g_sel,
      rho, beta_ge, beta_ge_sel,
      b_matrix, eigen_vec, eigen_value, p_intensity, wci
      );
  }
  
}
#endif


void calc_perturb_ug(
  int ne, int nb, HKL *reflex, int flag_bs,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int num_cal_thick, double thickness[], double accel_kvol,
  int num_branch, int *nth_branch,
  int n_sel_exact, int n_sel_bethe, int *selinfo_nth_reflex, int *nth_g_sel,
  double *Sg, double *beta_g, double *beta_g_sel, double K_n, 
  complex **matrix_unperturb, complex **eigen_vec, complex *eigen_value,
  double ***p_intensity, WORK_CAL_INT *wci
  )
{
  /*
   * Calculation of perturbed intensities
   *   (for a incident direction)
   *  with vg selection
   *
   * *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
   */
  int  it, ig, ip, nr, nr_sel, err;
  complex  **p_vg, **pa_mat, **pb_mat, *p_eigen_val;
  complex  **p_eigen_vec, *p_boundary_cond, **eigen_vec_inv;
  
  if(num_perturb == 0){ return; }	/* No calculation */
  
  nr = ne + nb;
  nr_sel = n_sel_exact + n_sel_bethe;
  
  /* ----- copy pointers of work area wci ----- */
  eigen_vec_inv = wci->eigen_vec_inv;
  p_vg = wci->p_vg;	pa_mat = wci->pa_mat;	pb_mat = wci->pb_mat;
  p_eigen_vec = wci->p_eigen_vec;	p_eigen_val = wci->p_eigen_val;
  p_boundary_cond = wci->p_boundary_cond;
  
  /* ----- Inverse matrix of the unperturbed eigen vector ----- */
  dbg_mon("$$ eigen_vec_inv\n");
  /* eigen_vec --> eigen_vec_inv */
  c_matcopy2(n_sel_exact, eigen_vec, eigen_vec_inv);
  if ( (err = inv_c_mat( n_sel_exact, epsilon.inv, eigen_vec_inv ) ) != 0 ){
    mb_fprintf(stdout, "$$ error (calc_perturb_ug): inv_c_mat = %d.\n", err);
  }
  
  for(ip=0; ip<num_perturb; ip++){
    /* ----- Transfer perturb_vgs to p_vg ----- */
    dbg_mon("$$ transfer perturbed vg\n");
    pvg_to_vg(ip, nr, perturb_vgs, p_vg);
    /*
      Perturb: Vg selection  ==> make use of unperturbed selinfo
     */
    /* ----- Perturb: Matrix ----- */
    dbg_mon("$$ p-construct_matrix\n");
    construct_ug_mat_slct(
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
      accel_kvol, p_vg, Sg, beta_g, pa_mat
      );
    dbg_print_cmat(
      dbg.a_mat_dump_flag, "a_matrix: (perturb)", n_sel_exact, pa_mat
      );
    
    /* ----- Perturb: Generalized Bethe approximation ----- */
    dbg_mon("$$ p-g-bethe\n");
    if(flag_bs < 10){
      g_bethe( n_sel_exact, n_sel_bethe, pa_mat, pb_mat );
    }
    else if(flag_bs >= 10){
      simple_bethe( n_sel_exact, n_sel_bethe, pa_mat, pb_mat );
    }
    
    dbg_print_cmat(
      dbg.b_mat_dump_flag, "bethe_matrix: (perturb)", n_sel_exact, pb_mat
      );
    
    /* ----- Perturbation ----- */
    dbg_mon("$$ perturbation2\n");
    perturbation2(
      eigen_value, eigen_vec, eigen_vec_inv,
      n_sel_exact, matrix_unperturb, pb_mat,
      epsilon.degeneracy, epsilon.eigen, p_eigen_val, p_eigen_vec
      );
    /*
			perturbation(
				eigen_val, eigen_vec, eigen_vec_inv, ne, bethe_mat, pb_mat,
				epsilon.degeneracy, epsilon.eigen, p_eigen_val, p_eigen_vec
			);
     */
    /*
			eigen_val_perturbation(
				eigen_val, eigen_vec, eigen_vec_inv,
				ne, bethe_mat, pb_mat, p_eigen_val
			);
     */
    dbg_print_cvec(
      dbg.eigval_dump_flag,"perturbed eigen values:",n_sel_exact, p_eigen_val
      );
    dbg_print_cmat(
      dbg.eigvec_dump_flag,"perturbed eigen vectors",n_sel_exact, p_eigen_vec
      );
    
    /* ----- Perturb: boundary condition ----- */
    dbg_mon("$$ p-upper_b_cond\n");
    upper_b_cond( n_sel_exact, beta_g_sel, p_eigen_vec, p_boundary_cond );
    dbg_print_cvec(
      dbg.b_cond_dump_flag,"p-upper_b_cond:",n_sel_exact, p_boundary_cond
      );
    
    /* ----- Perturb: calc. of perturbed intensities ----- */
    dbg_mon("$$ p-intensities\n");
    for(ig=0; ig<num_cal_g; ig++){
      for(it=0; it<num_cal_thick; it++){
        p_intensity[ip][ig][it]
          = psi_to_intensity_limited_ug(
            nth_g_sel[ig], thickness[it], n_sel_exact,
            num_branch, nth_branch, K_n,
            beta_g_sel, p_eigen_val, p_eigen_vec, p_boundary_cond
            );
      }
    }
  }

  dbg_mon("$$perturbation end\n");
}


void calc_intensity_with_perturb_ug(
  R_HKL K_xy, double abs_K, BEAM_SEARCH_PAR beam_search,
  int ne, int nb, HKL *reflex, complex **vg,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
  double **intensity, double ***p_intensity, WORK_CAL_INT *wci
  )
{
  /*
   * Calculation of diffracted intensities and perturbed intensities
   *  with vg selection (for an incident beam direction)
   *   *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
   */
  int it, ig, nr, ind, err, num_branch, *nth_branch;
  int n_sel_exact, n_sel_bethe, *selinfo_nth_reflex, *nth_g_sel;
  double *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond;

  nr = ne + nb;
  /* ----- copy pointers from work area wci ----- */
  Sg  = wci->Sg;
  beta_g  = wci->beta_g;
  beta_g_sel = wci->beta_g_sel;
  a_matrix = wci->a_matrix;
  b_matrix = wci->b_matrix;
  eigen_vec = wci->eigen_vec;
  eigen_value = wci->eigen_value;
  boundary_cond = wci->boundary_cond;
  nth_branch = wci->nth_branch;
  nth_g_sel = wci->nth_g_sel;
  selinfo_nth_reflex = wci->selinfo_nth_reflex;
  /* ----- calc. of Sg and beta_g ----- */
  dbg_mon("$$ calc_sg_beta_g\n");
  K_z = calc_K_z( abs_K, K_xy, bv);
  K_n = calc_K_n( K_xy, K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
  calc_sg_beta_g(
    K_xy, K_z, K_n, abs_K, abs_incidence, incidence, abs_surf_norm, surf_norm,
    ne, nb, reflex, bv, Sg, beta_g
    );
  /* ----- vg selection ----- */
  dbg_mon("$$ vg_select\n");
  vg_select_info_ug(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, Sg, beam_search, incidence, accel_kvol
    );
  vg_select_trans_ug(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_g, beta_g_sel, num_cal_g, nth_g, nth_g_sel
    );
  /* ----- matrix ----- */
  dbg_mon("$$ construct_matrix\n");
  construct_ug_mat_slct(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    accel_kvol, vg, Sg, beta_g, a_matrix
    );
  dbg_print_cmat(dbg.a_mat_dump_flag, "a_matrix:", n_sel_exact, a_matrix);
  /* ----- Bethe approximation ----- */
  dbg_mon("$$ Bethe Approx.\n");
  if(beam_search.flag_bs < 10){	/* Generalized Bethe */
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){ /* Simple Bethe */
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  dbg_print_cmat(dbg.b_mat_dump_flag, "bethe_matrix:", n_sel_exact, b_matrix);
  /* ----- Eigen value problem (QR method) ----- */
  dbg_mon("$$ eigen values and vectors\n");
  c_matcopy2( n_sel_exact, b_matrix, a_matrix);	/* b_matrix --> a_matrix */	
  if(epsilon.diagonalize_func == 0){
    eispack_cg( n_sel_exact, epsilon.eigen, a_matrix, eigen_value, eigen_vec );
    /* a_matrix is broken by eispack_cg. */
  }
  else if(epsilon.diagonalize_func == 1){
    err = eigen_c_mat(
      n_sel_exact, epsilon.eigen, ind = 1, a_matrix, eigen_value, eigen_vec
      );
    if( err != 0){
      mb_fprintf(stdout, "$$ (calc_intensity_loop): eigen_c_mat = %d\n",err);
    }
  }
  else{
    mb_fprintf(
      stdout, 
      "$$ Invalid epsilon.diagonalize_func=%d. Abort.\n",
      epsilon.diagonalize_func
      );
  }
  dbg_print_cvec(dbg.eigval_dump_flag," eigen values:",n_sel_exact, eigen_value);
  dbg_print_cmat(dbg.eigvec_dump_flag," eigen vectors",n_sel_exact, eigen_vec);
  /* ----- Boundary condition ----- */
  dbg_mon("$$ upper_b_cond\n");
  upper_b_cond( n_sel_exact, beta_g_sel, eigen_vec, boundary_cond );
  dbg_print_cvec(dbg.b_cond_dump_flag,"upper_b_cond:",n_sel_exact, boundary_cond);
  /* ----- Branch selection (for reducing calculation time) ----- */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  /* ----- Calc. intensity ----- */
  dbg_mon("$$ calc_intensity\n");
  for(ig=0; ig<num_cal_g; ig++){
    for(it=0; it<num_cal_thick; it++){
      /* printf("$$ thickness[%d]=%lf\n",it,thickness[it]); */
      intensity[ig][it]
        = psi_to_intensity_limited_ug(
          nth_g_sel[ig], thickness[it], n_sel_exact,
          num_branch, nth_branch, K_n,
          beta_g_sel, eigen_value, eigen_vec, boundary_cond
          );
      /*printf("$$ intensity[%d][%d]=%lf\n", ig, it, intensity[ig][it]);*/
    }
  }
  /*
   * Perturbation
   *   using perturbed potential perturb_vgs[ip][i][j],
   *   eigen values and eigen vectors
   */
  if(num_perturb > 0){
    calc_perturb_ug(
      ne, nb, reflex, beam_search.flag_bs, num_perturb, perturb_vgs,
      num_cal_g, num_cal_thick, thickness,
      accel_kvol, num_branch, nth_branch,
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex, nth_g_sel,
      Sg, beta_g, beta_g_sel, K_n,
      b_matrix, eigen_vec, eigen_value, p_intensity, wci
      );
  }
}


void calc_intensity_ug_coherent(
  R_HKL K_xy, double abs_K, complex initial_phase,
  BEAM_SEARCH_PAR beam_search,
  int ne, int nb, HKL *reflex, complex **vg,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
  complex **amplitude, complex ***p_amp, WORK_CAL_INT *wci
  )
     /*
      * Calculation of coherent diffraction intensities
      *   (for an incident beam direction)
      *  with vg selection
      *
      * *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
      */
{
  int it, ig, nr, ind, err, num_branch, *nth_branch;
  int n_sel_exact, n_sel_bethe, *selinfo_nth_reflex, *nth_g_sel;
  double *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond;
  
  nr = ne + nb;
  
  /* ----- copy pointers of work area wci ----- */
  Sg      = wci->Sg;	beta_g  = wci->beta_g;	beta_g_sel  = wci->beta_g_sel;
  a_matrix = wci->a_matrix;	b_matrix = wci->b_matrix;
  eigen_vec = wci->eigen_vec;	eigen_value = wci->eigen_value;
  boundary_cond = wci->boundary_cond;
  nth_branch = wci->nth_branch;	nth_g_sel = wci->nth_g_sel;
  selinfo_nth_reflex = wci->selinfo_nth_reflex;
  
  /* ----- calc. of Sg and beta_g ----- */	dbg_mon("$$ calc_sg_beta_g\n");
  K_z = calc_K_z( abs_K, K_xy, bv);
  K_n = calc_K_n( K_xy, K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
  calc_sg_beta_g(
    K_xy, K_z, K_n, abs_K, abs_incidence, incidence, abs_surf_norm, surf_norm,
    ne, nb, reflex, bv, Sg, beta_g
    );
  
  /* ----- vg selection ----- */	dbg_mon("$$ vg_select\n");
  vg_select_info_ug(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, Sg, beam_search, incidence, accel_kvol
    );
  vg_select_trans_ug(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_g, beta_g_sel, num_cal_g, nth_g, nth_g_sel
    );
  
  /* ----- matrix ----- */
  dbg_mon("$$ construct_matrix\n");
  construct_ug_mat_slct(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    accel_kvol, vg, Sg, beta_g, a_matrix
    );
  dbg_print_cmat(dbg.a_mat_dump_flag, "a_matrix:", n_sel_exact, a_matrix);
  
  /* ----- Generalized Bethe approximation ----- */
  dbg_mon("$$ GBA\n");
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  dbg_print_cmat(dbg.b_mat_dump_flag, "bethe_matrix:", n_sel_exact, b_matrix);
  
  /* ----- Eigen value problem (QR method) ----- */
  dbg_mon("$$ eigen values and vectors\n");
  c_matcopy2( n_sel_exact, b_matrix, a_matrix);	/* b_matrix --> a_matrix */	
  if(epsilon.diagonalize_func == 0){
    eispack_cg( n_sel_exact, epsilon.eigen, a_matrix, eigen_value, eigen_vec );
    /* a_matrix is broken by eispack_cg. */
  }
  else if(epsilon.diagonalize_func == 1){
    err = eigen_c_mat(
      n_sel_exact, epsilon.eigen, ind = 1, a_matrix, eigen_value, eigen_vec
      );
    if( err != 0){
      mb_fprintf(stdout, "$$ (calc_intensity_loop): eigen_c_mat = %d\n",err);
    }
  }
  else{
    mb_fprintf(
      stdout, 
      "$$ Invalid epsilon.diagonalize_func=%d. Abort.\n",
      epsilon.diagonalize_func
      );
  }
  dbg_print_cvec(dbg.eigval_dump_flag," eigen values:",n_sel_exact, eigen_value);
  dbg_print_cmat(dbg.eigvec_dump_flag," eigen vectors",n_sel_exact, eigen_vec);
  
  /* ----- Boundary condition ----- */
  dbg_mon("$$ upper_b_cond\n");
  upper_b_cond_coherent(
    n_sel_exact, beta_g_sel, initial_phase, eigen_vec, boundary_cond
    );	/* <<<<< */
  dbg_print_cvec(dbg.b_cond_dump_flag,"upper_b_cond:",n_sel_exact, boundary_cond);
  
  /* ----- Branch selection (for reduction of calc. time) ----- */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  
  /* ----- Calc. intensity ----- */
  dbg_mon("$$ calc_intensity\n");
  for(ig=0; ig<num_cal_g; ig++){
    for(it=0; it<num_cal_thick; it++){
      /* printf("$$ thickness[%d]=%lf\n",it,thickness[it]); */
      amplitude[ig][it]
        = psi_limited_ug(
          nth_g_sel[ig], thickness[it], n_sel_exact,
          num_branch, nth_branch, K_n,
          beta_g_sel, eigen_value, eigen_vec, boundary_cond
          );
    }
  }

  if(num_perturb > 0){
    mb_fprintf(
      stdout, 
      "$$ perturbation is not supported in coherent mode. Sorry.\n");
    exit(EXIT_FAILURE);
  }
#if 0
  /*
   * Perturbation
   *   using perturbed potential perturb_vgs[ip][i][j],
   *   eigen values and eigen vectors
   */
  
  if(num_perturb > 0){
    calc_perturb_ug(
      ne, nb, reflex, num_perturb, perturb_vgs,
      num_cal_g, num_cal_thick, thickness,
      accel_kvol, num_branch, nth_branch,
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex, nth_g_sel,
      Sg, beta_g, beta_g_sel, K_n,
      b_matrix, eigen_vec, eigen_value, p_intensity, wci
      );
  }
#endif
}


void calc_intensity_ug_spcell(
  R_HKL K_xy, double abs_K, complex *probe, BEAM_SEARCH_PAR beam_search,
  int ne, int nb, HKL *reflex, complex **vg,
  int num_perturb, complex ***perturb_vgs,
  int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
  double **intensity,	double ***p_intensity, WORK_CAL_INT *wci
  )
{
  /*
   * Calculation of diffracted intensities and perturbed intensities
   *   (for a incident direction)
   *  with vg selection
   *
   * *wci: work area (c.f.: alloc_work_cal_int & free_work_cal_int)
   */
  int it, ig, nr, ind, err, num_branch, *nth_branch;
  int n_sel_exact, n_sel_bethe, *selinfo_nth_reflex, *nth_g_sel;
  double *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond;
  
  nr = ne + nb;
  
  /* ----- copy pointers of work area wci ----- */
  Sg      = wci->Sg;	beta_g  = wci->beta_g;	beta_g_sel  = wci->beta_g_sel;
  a_matrix = wci->a_matrix;	b_matrix = wci->b_matrix;
  eigen_vec = wci->eigen_vec;	eigen_value = wci->eigen_value;
  boundary_cond = wci->boundary_cond;
  nth_branch = wci->nth_branch;	nth_g_sel = wci->nth_g_sel;
  selinfo_nth_reflex = wci->selinfo_nth_reflex;
  
  /* ----- calc. of Sg and beta_g ----- */	dbg_mon("$$ calc_sg_beta_g\n");
  K_z = calc_K_z( abs_K, K_xy, bv);
  K_n = calc_K_n( K_xy, K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
  calc_sg_beta_g(
    K_xy, K_z, K_n, abs_K, abs_incidence, incidence, abs_surf_norm, surf_norm,
    ne, nb, reflex, bv, Sg, beta_g
    );
  
  /* ----- vg selection ----- */	dbg_mon("$$ vg_select\n");
  vg_select_info_ug(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, Sg, beam_search, incidence, accel_kvol
    );
  vg_select_trans_ug(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_g, beta_g_sel, num_cal_g, nth_g, nth_g_sel
    );
  
  /* ----- matrix ----- */
  dbg_mon("$$ construct_matrix\n");
  construct_ug_mat_slct(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    accel_kvol, vg, Sg, beta_g, a_matrix
    );
  dbg_print_cmat(dbg.a_mat_dump_flag, "a_matrix:", n_sel_exact, a_matrix);
  
  /* ----- Generalized Bethe approximation ----- */
  dbg_mon("$$ GBA\n");
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  dbg_print_cmat(dbg.b_mat_dump_flag, "bethe_matrix:", n_sel_exact, b_matrix);
  
  /* ----- Eigen value problem (QR method) ----- */
  dbg_mon("$$ eigen values and vectors\n");
  c_matcopy2( n_sel_exact, b_matrix, a_matrix);	/* b_matrix --> a_matrix */	
  if(epsilon.diagonalize_func == 0){
    eispack_cg( n_sel_exact, epsilon.eigen, a_matrix, eigen_value, eigen_vec );
    /* a_matrix is broken by eispack_cg. */
  }
  else if(epsilon.diagonalize_func == 1){
    err = eigen_c_mat(
      n_sel_exact, epsilon.eigen, ind = 1, a_matrix, eigen_value, eigen_vec
      );
    if( err != 0){
      mb_fprintf(
        stdout, "$$ (calc_intensity_loop): eigen_c_mat = %d\n",err); }
  }
  else{
    mb_fprintf( stdout, 
      "$$ Invalid epsilon.diagonalize_func=%d. Abort.\n",
      epsilon.diagonalize_func
      );
  }
  dbg_print_cvec(dbg.eigval_dump_flag," eigen values:",n_sel_exact, eigen_value);
  dbg_print_cmat(dbg.eigvec_dump_flag," eigen vectors",n_sel_exact, eigen_vec);
  
  /* ----- Boundary condition ----- */
  dbg_mon("$$ upper_b_cond\n");
  upper_b_cond_spcell( n_sel_exact, beta_g_sel, probe, eigen_vec, boundary_cond );
  dbg_print_cvec(dbg.b_cond_dump_flag,"upper_b_cond:",n_sel_exact, boundary_cond);
  
  /* ----- Branch selection (for reduction of calc. time) ----- */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  
  /* ----- Calc. intensity ----- */
  dbg_mon("$$ calc_intensity\n");
  for(ig=0; ig<num_cal_g; ig++){
    for(it=0; it<num_cal_thick; it++){
      /* printf("$$ thickness[%d]=%lf\n",it,thickness[it]); */
      intensity[ig][it]
        = psi_to_intensity_limited_ug(
          nth_g_sel[ig], thickness[it], n_sel_exact,
          num_branch, nth_branch, K_n,
          beta_g_sel, eigen_value, eigen_vec, boundary_cond
          );
      /*printf("$$ intensity[%d][%d]=%lf\n", ig, it, intensity[ig][it]);*/
    }
  }
  
  /*
   * Perturbation
   *   using perturbed potential perturb_vgs[ip][i][j],
   *   eigen values and eigen vectors
   */
  
  if(num_perturb > 0){
    calc_perturb_ug(
      ne, nb, reflex, beam_search.flag_bs,
      num_perturb, perturb_vgs,
      num_cal_g, num_cal_thick, thickness,
      accel_kvol, num_branch, nth_branch,
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex, nth_g_sel,
      Sg, beta_g, beta_g_sel, K_n,
      b_matrix, eigen_vec, eigen_value, p_intensity, wci
      );
  }
  
}


void calc_disp_surf_ug(
  int ne, int nb, int num_cal_points, 
  double accel_kvol, LAT_PAR lp, CELL_PAR bv, double *H, double *G,
  R_HKL *K_xy, double abs_K, UVW incidence, UVW surf_norm, 
  HKL *reflex, complex **vg, int branch_begin, int branch_end, 
  int *rf_amp, int num_rf_amp, BEAM_SEARCH_PAR beam_search, 
  int num_thick, double *thickness, char *infile
  )
{
  int ibranch, ik, nr, i, ithick, prog, n_sel_exact, n_sel_bethe;
  int num_branch, *nth_branch, *nth_g_sel, *selinfo_nth_reflex;
  double intensity;
  double abs_incidence, abs_surf_norm, cosfai_inc_surf;
   double *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond, c_g;
  FILE *fo_ds, *fo_amp, *fo_int, *fo_mbutl;
  char outfile_ds[BUFLEN], outfile_amp[BUFLEN], outfile_int[BUFLEN];
  char outfile_mbutl[BUFLEN];

  epsilon.vg_select_flag = 0;
  mb_fprintf(stdout, "(calc_disp_surf_ug) vg_select_flag is set to 0.\n");

  /* ----- dispersion surface output ----- */
  sprintf(outfile_ds, "%s_ds.txt", infile);
  fo_ds = mb_fopen_chk_wmode(outfile_ds);

  mb_fprintf(fo_ds, "# Dispersion surface\n");
  mb_fprintf(fo_ds, "# kn, excitation(=eps^2)\n");

  /* ----- For Amplitude output ----- */
  sprintf(outfile_amp, "%s_amp.txt", infile);
  fo_amp = mb_fopen_chk_wmode(outfile_amp);

  mb_fprintf(
    fo_amp, "# Amplitudes of reflections Cg (multiplied by excitation)\n"
    );
  mb_fprintf(
    fo_amp, "# (Amplitude and Phase (rad),  g:reflection, j:branch)\n"
    );
  for(i=0; i<num_rf_amp; i++){
    mb_fprintf(
      fo_amp, "# g %d: %d %d %d\n",
      i, reflex[rf_amp[i]].h, reflex[rf_amp[i]].k, reflex[rf_amp[i]].l
      );
  }
  
  /* ----- For intensity output ----- */
  sprintf(outfile_int, "%s_int.txt", infile);
  fo_int = mb_fopen_chk_wmode(outfile_int);

  mb_fprintf(fo_int, "# Diffracted intensities\n");
  for(i=0; i<num_rf_amp; i++){
    mb_fprintf(
      fo_int, "# g %d: %d %d %d\n",
      i, reflex[rf_amp[i]].h, reflex[rf_amp[i]].k, reflex[rf_amp[i]].l
      );
  }
  for(ithick=0; ithick<num_thick; ithick++){
    mb_fprintf( fo_int, "# thickness %d: %lf[A]\n", ithick, thickness[ithick]);
  }
  
  nr = ne + nb;
  Sg      = new_vector( double, nr );			chk_vector(Sg);
  beta_g  = new_vector( double, nr );			chk_vector(beta_g);
  beta_g_sel  = new_vector( double, nr );		chk_vector(beta_g_sel);
  a_matrix = new_matrix( complex, nr, nr );	chk_matrix(a_matrix);
  b_matrix = new_matrix( complex, nr, nr );	chk_matrix(b_matrix);
  eigen_vec = new_matrix( complex, nr, nr );	chk_matrix(eigen_vec);
  eigen_value = new_vector( complex, nr );	chk_vector(eigen_value);
  boundary_cond = new_vector( complex, nr );	chk_vector(boundary_cond);
  nth_branch = new_vector(int, nr);			chk_vector(nth_branch);
  nth_g_sel = new_vector(int, num_rf_amp);	chk_vector(nth_g_sel);
  selinfo_nth_reflex = new_vector(int, nr);	chk_vector(selinfo_nth_reflex);
  
  /*	wave_length = wavelength(accel_kvol);*/
  abs_incidence = abs_uvw(incidence, bv);	/* incidence is upwards */
  abs_surf_norm = abs_uvw(surf_norm, bv);	/* surf_norm is upwards */
  cosfai_inc_surf
    = inner_product_uvw( incidence, surf_norm, bv )
      / ( abs_incidence * abs_surf_norm );
  
/*  mb_fprintf(fo_ds, "kxy\tKz");
  mb_fprintf(fo_amp, "kxy");
  mb_fprintf(fo_int, "kxy"); */
  mb_fprintf(fo_ds, "kxy(H)\tkxy(G)\tKz");
  mb_fprintf(fo_amp, "kxy(H)\tkxy(G)");
  mb_fprintf(fo_int, "kxy(H)\tkxy(G)");
  for(ibranch=branch_begin; ibranch<=branch_end; ibranch++){
    mb_fprintf(
      fo_ds, "\tRe(kz%d)\tIm(kz%d)\tRe(gamma%d)\tIm(gamma%d)\tSq(eps%d)",
      ibranch, ibranch, ibranch, ibranch, ibranch
      );
  }
  for(i=0; i<num_rf_amp; i++){
    for(ibranch=branch_begin; ibranch<=branch_end; ibranch++){
      mb_fprintf(
        fo_amp, "\tA(eC)g%dj%d\tP(eC)g%dj%d",i, ibranch, i, ibranch
        );
    }
    for(ithick=0; ithick<num_thick; ithick++){
      mb_fprintf(fo_int, "\tg%dt%d", i, ithick);
    }
  }
  mb_fprintf(fo_ds, "\n");
  mb_fprintf(fo_amp, "\n");
  mb_fprintf(fo_int, "\n");
  
  /*  prog = 0; */
  for(ik=0; ik<num_cal_points; ik++){

    progress_indicator(ik, num_cal_points);
    /* exec_progress(&prog, ik, num_cal_points); */

    /* calc. of Sg and beta_g */
    /* dbg_mon("$$ calc_sg_beta_g\n"); */
    K_z = calc_K_z( abs_K, K_xy[ik], bv);
    K_n = calc_K_n( K_xy[ik], K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
    calc_sg_beta_g(
      K_xy[ik], K_z, K_n, abs_K,
      abs_incidence, incidence, abs_surf_norm, surf_norm,
      ne, nb, reflex, bv, Sg, beta_g
      );

    /* ----- vg selection ----- */	dbg_mon("$$ vg_select\n");
    vg_select_info_ug(
      ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
      vg, reflex, Sg, beam_search, incidence, accel_kvol
      );
    vg_select_trans_ug(
      ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
      beta_g, beta_g_sel, num_rf_amp, rf_amp, nth_g_sel
      );
    construct_ug_mat_slct(
      n_sel_exact, n_sel_bethe, selinfo_nth_reflex,		
      accel_kvol, vg, Sg, beta_g, a_matrix
      );
    if(beam_search.flag_bs < 10){
      g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
    }
    else if(beam_search.flag_bs >= 10){
      simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
    }

    /* eigen_c_mat( ne, epsilon.eigen, 1, b_matrix, eigen_value, eigen_vec );*/
    eispack_cg(
      n_sel_exact, epsilon.eigen, b_matrix, eigen_value, eigen_vec
      );
    upper_b_cond( n_sel_exact, beta_g_sel, eigen_vec, boundary_cond );

    /* ----- Branch selection (for reduction of calc. speed) ----- */
    limited_branch(
      epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
      );

    /* mb_fprintf(fo_ds, "%d\t%.12lg", ik, fabs(K_n)); */
    mb_fprintf(fo_ds, "%lf\t%lf\t%.12lg", H[ik], G[ik], fabs(K_n)); /* <<<<< */

    /* Caution: branch 1 corresponds to eigenstate[0]. */
    for(ibranch=branch_begin; ibranch<=branch_end; ibranch++){
      mb_fprintf(
        fo_ds, "\t%.12lg\t%lg\t%lg\t%lg\t%lg",
        fabs( K_n + eigen_value[ibranch - 1].re / (2.0 * K_n) ),
        fabs( eigen_value[ibranch - 1].im / (2.0 * K_n) ),
        eigen_value[ibranch - 1].re,
        eigen_value[ibranch - 1].im,
        sq(boundary_cond[ibranch - 1].re) + sq(boundary_cond[ibranch - 1].im)
        );
    }
   /* mb_fprintf(fo_amp,"%d", ik);
    mb_fprintf(fo_int,"%d", ik); */
    mb_fprintf(fo_amp,"%lf\t%lf", H[ik], G[ik]);
    mb_fprintf(fo_int,"%lf\t%lf", H[ik], G[ik]);
    for(i=0; i<num_rf_amp; i++){
      
      for(ithick=0; ithick<num_thick; ithick++){
        intensity
          = psi_to_intensity_limited_ug(
            nth_g_sel[i], thickness[ithick], n_sel_exact,
            num_branch, nth_branch, K_n,
            beta_g_sel, eigen_value, eigen_vec, boundary_cond
            )
            * epsilon.intensity_scaling;
        mb_fprintf( fo_int, "\t%lf", intensity);
      }

      for(ibranch=branch_begin; ibranch<=branch_end; ibranch++){
        c_g = c_div(
          /*  bc * c_g  */
          c_mul(
            boundary_cond[ibranch - 1], eigen_vec[rf_amp[i]][ibranch - 1]
            ),
          c_conv(sqrt(beta_g[rf_amp[i]]), 0)
          );
        mb_fprintf(
          fo_amp, "\t%lg\t%lg", c_abs(c_g), c_arg(c_g)
          ); /* amplitude */
      }

    }
    mb_fprintf(fo_ds, "\n");
    mb_fprintf(fo_amp, "\n");
    mb_fprintf(fo_int, "\n");

  }
  mb_fprintf(stdout, "100%%\n");
  mb_fclose(fo_ds);
  mb_fclose(fo_amp);
  mb_fclose(fo_int);

  /* ----- For mbutl output ----- */
  sprintf(outfile_mbutl, "%s_mbutl.txt", infile);
  fo_mbutl = mb_fopen_chk_wmode(outfile_mbutl);
  mb_fprintf(fo_mbutl, "MBUTL_SCRIPT\n");
  mb_fprintf(fo_mbutl, "DISPSURF_VISUAL_MODE\n");
  mb_fprintf(fo_mbutl, "HTML_OPEN\n");
  mb_fprintf(fo_mbutl, "%s.html\n", infile);
  mb_fprintf(fo_mbutl, "DISPSURF_VISUAL\n");
  mb_fprintf(fo_mbutl, "%d  #branch_begin\n", branch_begin);
  mb_fprintf(fo_mbutl, "%d  #branch_end\n", branch_end);
  mb_fprintf(fo_mbutl, "%d  #num_rf_amp\n", num_rf_amp);
  mb_fprintf(fo_mbutl, "%d  #num_thick\n", num_thick);
  mb_fprintf(fo_mbutl, "%s\n", infile);
  mb_fprintf(fo_mbutl, "%s\n", outfile_ds);
  mb_fprintf(fo_mbutl, "%s\n", outfile_amp);
  mb_fprintf(fo_mbutl, "%s\n", outfile_int);
  mb_fprintf(fo_mbutl, "HTML_CLOSE\n");
  mb_fclose(fo_mbutl);
  mb_fprintf(stdout, "\n==> mbutl %s_mbutl.txt\n", infile);

  free_vector(Sg);
  free_vector(beta_g);
  free_vector(beta_g_sel);
  free_matrix(a_matrix);
  free_matrix(b_matrix);
  free_matrix(eigen_vec);
  free_vector(eigen_value);
  free_vector(boundary_cond);
  free_vector(nth_branch);
  free_vector(nth_g_sel);
  free_vector(selinfo_nth_reflex);
}


complex calc_projpot( R_UVW rx, int nr, HKL reflex[], complex **vg )
{
  /*
   * Calculate projected potential
   */
  int ig;
  complex v;
  double gr;
  
  v = c_conv(0.0,0.0);
  for(ig=0; ig<nr; ig++){
    gr	= (double) reflex[ig].h * rx.u
      + (double) reflex[ig].k * rx.v
        + (double) reflex[ig].l * rx.w;
    v = c_add(
      v,
      c_mul( vg[ig][0], c_exp( c_conv(0.0, PI * 2.0 * gr) ) )
      );
  }
  return v;
}


complex calc_inv_fourier( R_UVW rx, int nr, HKL reflex[], complex *vg )
{
  /*
    Calculate density distribution using inverse Frourier transform
   */
  int ig;
  complex v;
  double gr;
  
  v = c_conv(0.0,0.0);
  for(ig=0; ig<nr; ig++){
    gr = (double) reflex[ig].h * rx.u
      + (double) reflex[ig].k * rx.v
        + (double) reflex[ig].l * rx.w;
    v = c_add( v, c_mul(vg[ig], c_exp(c_conv(0.0, PI * 2.0 * gr))) );
  }
  return v;
}


complex calc_bloch_state_func_ug(
  int nth_branch, R_UVW rx, int num_exact, HKL reflex[], R_HKL k_xy,
  double *beta_g, complex *eigen_value, complex **eigen_vec
  )
{
  /*
   * Calculate wave function of n-th Bloch state
   */
  int ig;
  complex phi;
  double kr, gr;
  
  phi = c_conv(0.0,0.0);
  for(ig=0; ig<num_exact; ig++){
    gr	= (double) reflex[ig].h * rx.u
      + (double) reflex[ig].k * rx.v
        + (double) reflex[ig].l * rx.w;
    phi = c_add(
      phi,
      c_mul(
        c_div(
          eigen_vec[ig][nth_branch],
          c_conv( sqrt(beta_g[ig]), 0.0 )
          ),
        c_exp( c_conv(0.0, PI * 2.0 * gr) )
        )
      );
  }
  /*printf("$$ k_xy=%lf %lf %lf\n", k_xy.h,k_xy.k,k_xy.l);*/
  kr = k_xy.h * rx.u + k_xy.k * rx.v + k_xy.l * rx.w;
  phi = c_mul(phi, c_exp( c_conv(0.0, PI * 2.0 * kr)) );
  return phi;
}


void calc_bloch_state_ug(
  double accel_kvol, LAT_PAR lp, CELL_PAR bv,
  R_HKL k_xy, double abs_K, UVW incidence, UVW surf_norm,
  int num_exact, int num_bethe, HKL reflex[], complex **vg,
  BEAM_SEARCH_PAR beam_search,
  double x_begin, double x_end, double y_begin, double y_end,
  R_UVW x1, R_UVW x2, int num_points_x, int num_points_y,
  int branch_begin, int branch_end,
  char *infile, FILE *fo_log
  )
{
  int  nr, ip, ix, iy, ib, nmx, nmy, num_branch, prog, *nth_branch, *cal_g;
  int  n_sel_exact, n_sel_bethe, num_cal_g=0, *nth_g_sel, *selinfo_nth_reflex;
  int image_width,image_height;
  float *data;
  double  *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  double  abs_incidence, abs_surf_norm, cosfai_inc_surf;
  double  bs_density, x_scale, y_scale, excitation;
  /* double  wave_length, e_density, ev_dash; */
  complex  **a_matrix, **b_matrix, **eigen_vec, *eigen_value, *boundary_cond;
  complex  projp, bs_func;
  char  kn_outfile[BUFLEN], bl_outfile[BUFLEN], bl_phase_outfile[BUFLEN],tiff_out[BUFLEN];
  char outfile_mbutl[BUFLEN];
  R_UVW  rx;
  FILE  *fo, *fo_mbutl;
  
  epsilon.vg_select_flag = 0;
  nr = num_exact + num_bethe;
  Sg      = new_vector( double, nr );			chk_vector(Sg);
  beta_g  = new_vector( double, nr );			chk_vector(beta_g);
  beta_g_sel  = new_vector( double, nr );		chk_vector(beta_g_sel);
  a_matrix = new_matrix( complex, nr, nr );	chk_matrix(a_matrix);
  b_matrix = new_matrix( complex, nr, nr );	chk_matrix(b_matrix);
  eigen_vec = new_matrix( complex, nr, nr );	chk_matrix(eigen_vec);
  eigen_value = new_vector( complex, nr );	chk_vector(eigen_value);
  boundary_cond = new_vector( complex, nr );	chk_vector(boundary_cond);
  num_branch = branch_end - branch_begin + 1;
  nth_branch = new_vector(int, nr);			chk_vector(nth_branch);
  /*	nth_g_sel = new_vector(int, num_rf_amp);	chk_vector(nth_g_sel); */
  selinfo_nth_reflex = new_vector(int, nr);	chk_vector(selinfo_nth_reflex);
  
  abs_incidence = abs_uvw(incidence, bv);  /* incidence is upwards */
  abs_surf_norm = abs_uvw(surf_norm, bv);  /* surf_norm is upwards */
  cosfai_inc_surf
    = inner_product_uvw( incidence, surf_norm, bv )
      / ( abs_incidence * abs_surf_norm );
  
  /* ----- calc. of Sg and beta_g ----- */
  /* dbg_mon("$$ calc_sg_beta_g\n"); */
  K_z = calc_K_z( abs_K, k_xy, bv);
  K_n = calc_K_n( k_xy, K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
  calc_sg_beta_g(
    k_xy, K_z, K_n, abs_K, abs_incidence, incidence, abs_surf_norm, surf_norm,
    num_exact, num_bethe, reflex, bv, Sg, beta_g
    );
  
  /* ----- vg selection ----- */
  dbg_mon("$$ vg_select\n");
  vg_no_select_info_ug(
    num_exact, num_bethe, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex
    );
  num_cal_g=0;
  vg_select_trans_ug(
    num_exact, num_bethe, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_g, beta_g_sel, num_cal_g, cal_g, nth_g_sel
    );
  
  construct_ug_mat_slct(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    accel_kvol, vg, Sg, beta_g, a_matrix
    );
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  
  /* eigen_c_mat( ne, epsilon.eigen, 1, b_matrix, eigen_value, eigen_vec );*/
  eispack_cg( n_sel_exact, epsilon.eigen, b_matrix, eigen_value, eigen_vec );
  upper_b_cond( n_sel_exact, beta_g_sel, eigen_vec, boundary_cond );
  
  limited_branch(  /* Branch selection (for reducing calc. time) */
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );
  
  /*
   * Result output
   */
  /* ----- mbutl output 1 ----- */
  sprintf(outfile_mbutl, "%s_mbutl.txt", infile);
  fo_mbutl = mb_fopen_chk_wmode(outfile_mbutl);
  mb_fprintf(fo_mbutl, "MBUTL_SCRIPT\n");
  mb_fprintf(fo_mbutl, "BLOCH_VISUAL_MODE\n");
  mb_fprintf(fo_mbutl, "HTML_OPEN\n");
  mb_fprintf(fo_mbutl, "%s.html\n", infile);
  mb_fprintf(fo_mbutl, "BLOCH_VISUAL\n");
  mb_fprintf(fo_mbutl, "%s\n", infile);
  
  /* ----- kn file ----- */
  sprintf(kn_outfile, "%s_kn.txt", infile);
  fo = mb_fopen_chk_wmode(kn_outfile);

  mb_fprintf(fo, "%% kn\n");
  /* fprintf(fo, "%% i: Re{kz}  Im{kz}  eps^2\n"); */
  mb_fprintf(fo, "%% branch_i\tRe_eigenvalue\tIm_eigenvalue\teps^2\n");
  for(ib=0; ib<n_sel_exact; ib++){
    excitation = c_abs(boundary_cond[ib]);
    mb_fprintf(fo, "%d: %lf %lf %lf\n", ib+1,
            eigen_value[ib].re, eigen_value[ib].im, sq(excitation)); 
  }
  mb_fclose(fo);
  mb_fprintf(fo_log, " k_n ==> %s\n", kn_outfile);
  
  /*
   * Projected potential
   */
  sprintf(bl_outfile, "%s_pp.txt", infile);
  fo = mb_fopen_chk_wmode(bl_outfile);
  mb_fprintf(fo, "# Projected potential");
  mb_fprintf(fo, " (formed by %d-beams)\n", num_exact + num_bethe);
  mb_fprintf(fo, "#nx ny\n");
  mb_fprintf(fo, "#%d %d\n", num_points_x, num_points_y);
  mb_fprintf(fo, "#x\ty\t\tpotential.re[V]\tpotential.im[V]\n");

  nmx = num_points_x - 1;
  nmy = num_points_y - 1;

  mb_fprintf(stdout, "Proj. potential\n");
  for(ix=0; ix<num_points_x; ix++){
    progress_indicator(ix, num_points_x);
    if(nmx == 0){ x_scale = x_begin; }
    else{ x_scale = x_begin + (x_end - x_begin) * (double) ix / (double) nmx; }
    for(iy=0; iy<num_points_y; iy++){
      if(nmy == 0){ y_scale = y_begin; }
      else{ y_scale = y_begin + (y_end - y_begin) * (double) iy / (double) nmy; }
      rx.u = x_scale * x1.u + y_scale * x2.u;
      rx.v = x_scale * x1.v + y_scale * x2.v;
      rx.w = x_scale * x1.w + y_scale * x2.w;
      projp = calc_projpot( rx, num_exact + num_bethe, reflex, vg );
      /* e_density = calc_e_density(  ); */
      mb_fprintf(
        fo, "%g\t%g\t%g\t%g\n", x_scale, y_scale, projp.re, projp.im
        );
    }
    mb_fprintf(fo, "\n");
  }
  mb_fprintf(stdout, "100%%.\n");
  mb_fclose(fo);
  mb_fprintf( fo_log, " Projected potential: ==> %s\n", bl_outfile );
  mb_fprintf(fo_mbutl, "%s_pp.txt\n", infile);
  
  /*
   * Bloch state density file
   */
  mb_fprintf(fo_mbutl, "%d    # branch_begin\n", branch_begin);
  mb_fprintf(fo_mbutl, "%d    # branch_end\n", branch_end);
  for(ib=branch_begin; ib<=branch_end; ib++){
    
    sprintf(bl_outfile, "%s_b%d.txt", infile, ib);
    fo = mb_fopen_chk_wmode(bl_outfile);
    mb_fprintf(
      fo_log, " Prob. density (branch %d): ==> %s\n", ib, bl_outfile
      );
    mb_fprintf(fo_mbutl, "%s\n", bl_outfile);
    mb_fprintf(fo_mbutl, "%d    # branch\n", ib);
    mb_fprintf(fo_mbutl, "%g    # eps^2\n", sq(c_abs(boundary_cond[ib-1])));

    mb_fprintf(fo, "# Bloch state: branch %d)\n", ib);
    excitation = c_abs(boundary_cond[ib-1]);
    mb_fprintf(
      fo, "# branch:%d, eigen_val:%lf, excitation(eps^2):%lf\n", ib,
      eigen_value[ib-1].re, sq(excitation)
      );
    mb_fprintf(fo, "# nx ny\n");
    mb_fprintf(fo, "# %d %d\n", num_points_x, num_points_y);
    mb_fprintf(fo, "#x\ty\tdensity\tangle\tre\tim\n");

    nmx = num_points_x - 1;
    nmy = num_points_y - 1;

    mb_fprintf(stdout, "Branch %d\n", ib);

    image_width=num_points_x;
    image_height=num_points_y;
    data=malloc(image_width*image_height*sizeof(float));
    for(ip=0;ip<image_width*image_height;ip++){
       *(data+ip)=0.0 ;
    }
        
    for(ix=0; ix<num_points_x; ix++){
      progress_indicator(ix, num_points_x);
      if(nmx == 0){ x_scale = x_begin; }
      else{ x_scale = x_begin + (x_end - x_begin) * (double) ix / (double) nmx; }
      for(iy=0; iy<num_points_y; iy++){
        if(nmy == 0){ y_scale = y_begin; }
        else{ y_scale = y_begin + (y_end - y_begin) * (double) iy / (double) nmy; }
        rx.u = x_scale * x1.u + y_scale * x2.u;
        rx.v = x_scale * x1.v + y_scale * x2.v;
        rx.w = x_scale * x1.w + y_scale * x2.w;
        /* Caution: branch ib ==> element (ib-1) */
        bs_func
          = calc_bloch_state_func_ug(
            ib-1, rx, num_exact, reflex, k_xy, beta_g, eigen_value, eigen_vec
            );
        bs_density = sq(bs_func.re) + sq(bs_func.im);
        mb_fprintf(
          fo,  "%g\t%g\t%g\t%g\t%g\t%g\n",
          x_scale, y_scale, bs_density, c_arg(bs_func), bs_func.re, bs_func.im
          );
        *(data+iy+ix*image_width)=bs_density;
      }
      mb_fprintf(fo, "\n");
    }
    mb_fprintf(stdout, "100%%.\n");
    mb_fclose(fo);
    sprintf(tiff_out,"%s_b%d.tif",infile,ib);
    WriteImageDataToByteTiff(tiff_out,data,image_width,image_height,0) ;
    free(data);
  }

  /* ----- mbutl output ----- */
  mb_fprintf(fo_mbutl, "HTML_CLOSE\n");
  mb_fclose(fo_mbutl);
  mb_fprintf(stdout, "\n==> mbutl %s_mbutl.txt\n", infile);
  
  free_vector(Sg);
  free_vector(beta_g);
  free_vector(beta_g_sel);
  free_matrix(a_matrix);  free_matrix(b_matrix);
  free_matrix(eigen_vec);  free_vector(eigen_value);
  free_vector(boundary_cond);
  free_vector(nth_branch);
  if(num_cal_g > 0) {free_vector(nth_g_sel);}
  free_vector(selinfo_nth_reflex);
}


/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
void calc_tfringe_ug(
  double accel_kvol, LAT_PAR lp, CELL_PAR bv, R_HKL K_xy, 
  double abs_K, UVW incidence, UVW surf_norm, 
  int ne, int nb, HKL *reflex, complex **vg,
  BEAM_SEARCH_PAR beam_search,
  int num_cal_g, int *nth_g, int num_thick, double *thickness, char *infile
  )
{
  int ik, nr, i, it, n_sel_exact, n_sel_bethe;
  int num_branch, *nth_branch;
  int *nth_g_sel, *selinfo_nth_reflex;
  double intensity;
  double abs_incidence, abs_surf_norm, cosfai_inc_surf;
  double *Sg, *beta_g, *beta_g_sel, K_z, K_n;
  complex **a_matrix, **b_matrix;
  complex **eigen_vec, *eigen_value, *boundary_cond, c_g, psi;
  FILE *fo_int, *fo_plt;
  char int_file[BUFLEN], plt_file[BUFLEN];
  FILE *fo_re_im;
  double sqscl;
  char re_im_file[BUFLEN];

  sprintf(int_file, "%s_int.txt", infile);
  fo_int = mb_fopen_chk_wmode(int_file);
  sprintf(plt_file, "%s.plt", infile);
  fo_plt = mb_fopen_chk_wmode(plt_file);
  mb_fprintf(fo_plt, "#\n");
  sprintf(re_im_file, "%s_re_im.txt", infile);
  fo_re_im = mb_fopen_chk_wmode(re_im_file);

  nr = ne + nb;
  Sg      = new_vector( double, nr );			chk_vector(Sg);
  beta_g  = new_vector( double, nr );			chk_vector(beta_g);
  beta_g_sel  = new_vector( double, nr );		chk_vector(beta_g_sel);
  a_matrix = new_matrix( complex, nr, nr );	chk_matrix(a_matrix);
  b_matrix = new_matrix( complex, nr, nr );	chk_matrix(b_matrix);
  eigen_vec = new_matrix( complex, nr, nr );	chk_matrix(eigen_vec);
  eigen_value = new_vector( complex, nr );	chk_vector(eigen_value);
  boundary_cond = new_vector( complex, nr );	chk_vector(boundary_cond);
  nth_branch = new_vector(int, nr);			chk_vector(nth_branch);
  nth_g_sel = new_vector(int, num_cal_g);	chk_vector(nth_g_sel);
  selinfo_nth_reflex = new_vector(int, nr);	chk_vector(selinfo_nth_reflex);
  
  /* wave_length = wavelength(accel_kvol);*/
  abs_incidence = abs_uvw(incidence, bv);	/* incidence is upwards */
  abs_surf_norm = abs_uvw(surf_norm, bv);	/* surf_norm is upwards */
  cosfai_inc_surf =
    inner_product_uvw( incidence, surf_norm, bv )
      / ( abs_incidence * abs_surf_norm );

  /* calc. of Sg and beta_g */
  /* dbg_mon("$$ calc_sg_beta_g\n"); */
  K_z = calc_K_z( abs_K, K_xy, bv);
  K_n = calc_K_n( K_xy, K_z, abs_surf_norm, surf_norm, cosfai_inc_surf );
  calc_sg_beta_g(
    K_xy, K_z, K_n, abs_K,
    abs_incidence, incidence, abs_surf_norm, surf_norm,
    ne, nb, reflex, bv, Sg, beta_g
    );

    /* ----- vg selection ----- */	dbg_mon("$$ vg_select\n");
  vg_select_info_ug(
    ne, nb, &n_sel_exact, &n_sel_bethe, selinfo_nth_reflex,
    vg, reflex, Sg, beam_search, incidence, accel_kvol
    );
  vg_select_trans_ug(
    ne, nb, n_sel_exact, n_sel_bethe, selinfo_nth_reflex,
    beta_g, beta_g_sel, num_cal_g, nth_g, nth_g_sel
    );
  construct_ug_mat_slct(
    n_sel_exact, n_sel_bethe, selinfo_nth_reflex,	
    accel_kvol, vg, Sg, beta_g, a_matrix
    );
  if(beam_search.flag_bs < 10){
    g_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  else if(beam_search.flag_bs >= 10){
    simple_bethe( n_sel_exact, n_sel_bethe, a_matrix, b_matrix );
  }
  
  /* eigen_c_mat( ne, epsilon.eigen, 1, b_matrix, eigen_value, eigen_vec );*/
  eispack_cg(
    n_sel_exact, epsilon.eigen, b_matrix, eigen_value, eigen_vec
    );
  upper_b_cond( n_sel_exact, beta_g_sel, eigen_vec, boundary_cond );
  
  /* ----- Branch selection (for reduction of calc. speed) ----- */
  limited_branch(
    epsilon.bc_accuracy, n_sel_exact, boundary_cond, &num_branch, nth_branch
    );

  mb_fprintf(fo_int, "#thickness");
  mb_fprintf(fo_re_im, "#thickness");

  mb_fprintf(fo_plt, "set xlabel \"thickness [A]\"\n");
  mb_fprintf(fo_plt, "set ylabel \"Intensity\"\n");
  mb_fprintf(fo_plt, "set title \"Thickness fringes\"\n");

  mb_fprintf(fo_plt, "plot ");
  for(i=0; i<num_cal_g; i++){
    mb_fprintf(
      fo_int, "\t%d_%d_%d",
      reflex[nth_g[i]].h, reflex[nth_g[i]].k, reflex[nth_g[i]].l
      );
    mb_fprintf(
      fo_re_im, "\t%d_%d_%d",
      reflex[nth_g[i]].h, reflex[nth_g[i]].k, reflex[nth_g[i]].l
      );
    mb_fprintf(
      fo_plt, "\"%s\" using 1:%d title \"%d %d %d\" with lines ",
      int_file, i+2, reflex[nth_g[i]].h, reflex[nth_g[i]].k, reflex[nth_g[i]].l
      );
    if(i!=num_cal_g-1){
      mb_fprintf(fo_plt, ",\\\n");
    }
  }
  mb_fprintf(fo_int, "\n");
  mb_fprintf(fo_re_im, "\n");
  mb_fprintf(fo_plt, "\n");

  sqscl = sqrt(epsilon.intensity_scaling);
  for(it=0; it<num_thick; it++){
    mb_fprintf(fo_int, "%lg", thickness[it]);
    mb_fprintf(fo_re_im, "%lg", thickness[it]);
    for(i=0; i<num_cal_g; i++){
      intensity =
        psi_to_intensity_limited_ug(
          nth_g_sel[i], thickness[it], n_sel_exact,
          num_branch, nth_branch, K_n,
          beta_g_sel, eigen_value, eigen_vec, boundary_cond
          )
          * epsilon.intensity_scaling;
      mb_fprintf( fo_int, "\t%lg", intensity);
      psi = 
        psi_limited_ug(
          nth_g_sel[i], thickness[it], n_sel_exact,
          num_branch, nth_branch, K_n,
          beta_g_sel, eigen_value, eigen_vec, boundary_cond
          );
      mb_fprintf(
        fo_re_im, "\t%lg\t%lg", psi.re * sqscl, psi.im * sqscl
        );
    }
    mb_fprintf(fo_int, "\n");
    mb_fprintf(fo_re_im, "\n");
  }

  mb_fprintf(fo_plt, "pause -1 \n");
  mb_fclose(fo_int);
  mb_fclose(fo_plt);
  mb_fclose(fo_re_im);

  mb_fprintf(stdout, "\n==> SEE %s\n", plt_file);

  free_vector(Sg);
  free_vector(beta_g);
  free_vector(beta_g_sel);
  free_matrix(a_matrix);
  free_matrix(b_matrix);
  free_matrix(eigen_vec);
  free_vector(eigen_value);
  free_vector(boundary_cond);
  free_vector(nth_branch);
  free_vector(nth_g_sel);
  free_vector(selinfo_nth_reflex);
}
